from bfieldtools import mesh_conductor
from bfieldtools.utils import load_example_mesh

import pytest

import numpy as np
from numpy.testing import (
    assert_array_almost_equal,
    assert_array_equal,
    assert_allclose,
    assert_equal,
)


def _fake_mesh_conductor(mesh_name="10x10_plane", **kwargs):
    """
    Creates an example 'fake' MeshConductor object
    """
    return mesh_conductor.MeshConductor(mesh_obj=load_example_mesh(mesh_name), **kwargs)


def _fake_streamfunction(vals=None, mesh_name="unit_disc", **kwargs):
    """
    Creates an example 'fake' StreamFunction object
    """
    if vals:
        return mesh_conductor.StreamFunction(
            vals, _fake_mesh_conductor(mesh_name, **kwargs)
        )
    else:
        mesh = load_example_mesh(mesh_name)

        vals = 1 - np.linalg.norm(mesh.vertices, axis=1)

        return mesh_conductor.StreamFunction(
            vals, _fake_mesh_conductor(mesh_name, basis_name="vertex")
        )


def test_mesh_conductor_creation():
    """
    Tests different ways to create MeshConductor objects, check that basis operators work
    """
    for test_mesh in ["unit_sphere", "unit_disc", "plane_w_holes"]:
        for basis_name in ["suh", "inner", "vertex"]:

            c = _fake_mesh_conductor(
                mesh_name=test_mesh, basis_name=basis_name, N_suh=10
            )

            assert c.inner2vert.shape == (
                len(c.mesh.vertices),
                len(c.inner_vertices) + len(c.holes),
            )
            assert c.vert2inner.shape == (
                len(c.inner_vertices) + len(c.holes),
                len(c.mesh.vertices),
            )

            assert_array_almost_equal(
                (c.vert2inner @ c.inner2vert).toarray(),
                np.identity(len(c.inner_vertices) + len(c.holes)),
            )

            inner_diag = np.zeros((len(c.mesh.vertices), len(c.mesh.vertices)))
            inner_diag[c.inner_vertices, c.inner_vertices] = 1

            for hole in c.holes:
                inner_diag[np.asarray(hole)[:, None], np.asarray(hole)] += 1 / len(hole)

            assert_array_equal((c.inner2vert @ c.vert2inner).toarray(), inner_diag)

            if basis_name == "suh":
                assert c.basis.shape == (len(c.mesh.vertices), c.opts["N_suh"])


def test_mesh_conductor_resistance_update():

    obj = _fake_mesh_conductor(resistivity=1)

    R1 = obj.resistance

    obj.resistivity = 10

    assert obj.resistivity == 10

    R2 = obj.resistance

    assert_allclose(R2 / 10, R1)


def test_streamfunction_creation():
    """
    Test creating StreamFunctions with different bases
    """

    mesh_conductor.StreamFunction(
        np.zeros((10,)), _fake_mesh_conductor(basis_name="suh", N_suh=10)
    )
    mesh_conductor.StreamFunction(
        np.zeros((584,)), _fake_mesh_conductor(basis_name="inner")
    )
    mesh_conductor.StreamFunction(
        np.zeros((676,)), _fake_mesh_conductor(basis_name="vertex")
    )


def test_mesh_conductor_attributes():
    """
    tests MeshConductor attributes
    """
    c = _fake_mesh_conductor("unit_disc")

    c2 = _fake_mesh_conductor("unit_disc")
    c2.mesh.vertices += np.array([0, 0, 1])

    ind = c.inductance
    m_ind = c.mutual_inductance(c2)

    c2.sph_couplings

    c2.set_sph_options(N_sph=3)

    # This should cause an error
    try:
        c.set_basis("foo")
    except:
        print("Failed as expected")

    c.plot_mesh()

    U = c.U_coupling(np.array([[0, 0, 2]]))
    s = np.ones(c.basis.shape[1])
    b0 = U @ s
    c.U_coupling.reset()

    # Test multiplication inside the functions
    b1 = c.U_coupling(np.array([[0, 0, 2]]), s=s)

    assert_allclose(b0, b1)


def test_streamfunction_attributes():
    """
    tests StreamFunction functionality
    """
    s = _fake_streamfunction(mesh_name="unit_disc")
    s.__repr__()

    assert len(s.inner) == len(s.mesh_conductor.inner_vertices)
    assert len(s.vert) == len(s.mesh_conductor.mesh.vertices)

    s.power
    s.magnetic_energy

    s.plot()
    s.plot(contours=6)

    lp = s.discretize(N_contours=6)



"""
Logo generation
=========================

This script generates the bfieldtools logo
"""

#
# import numpy as np
# import matplotlib.pyplot as plt
# import sys
# path = '/m/home/home8/80/makinea1/unix/pythonstuff/bfieldtools'
# if path not in sys.path:
#    sys.path.insert(0,path)
#
# from bfieldtools.integrals import triangle_potential_dipole_linear
# from bfieldtools.integrals import omega
# from bfieldtools.utils import tri_normals_and_areas
# from bfieldtools.mesh_calculus import gradient
# from bfieldtools.mesh_magnetics import scalar_potential_coupling as compute_U
# from bfieldtools.mesh_magnetics import vector_potential_coupling as compute_A
# from bfieldtools.mesh_magnetics import magnetic_field_coupling as compute_C
from bfieldtools.mesh_magnetics import (
    magnetic_field_coupling_analytic as compute_C_analytic,
)

import trimesh
from mayavi import mlab
import numpy as np

#########################################################
# Test potential shape slightly above the surface

scalars = np.zeros(7)
scalars[0] = 1


def plot_element():
    x = np.sin(np.pi / 6)
    y = np.cos(np.pi / 6)
    points = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [x, y, 0],
            [-x, y, 0],
            [-1, 0, 0],
            [-x, -y, 0],
            [x, -y, 0],
        ]
    )

    tris = np.array([[0, 1, 2], [0, 2, 3], [0, 3, 4], [0, 4, 5], [0, 5, 6], [0, 6, 1]])
    mesh = trimesh.Trimesh(points, tris)
    # Stream function
    mlab.triangular_mesh(*points.T, tris, scalars=scalars, colormap="Blues")
    # Stream lines
    s2 = mlab.triangular_mesh(*points.T, tris, scalars=scalars, colormap="Blues")
    s2.enable_contours = True
    s2.actor.mapper.scalar_range = np.array([0.0, 1.0])
    s2.actor.mapper.scalar_visibility = False
    s2.actor.property.render_lines_as_tubes = True
    s2.actor.property.line_width = 3.0
    s2.contour.number_of_contours = 6
    s2.actor.property.color = (1.0, 0.0, 0.0)

    return s2, mesh


s, mesh = plot_element()

#%%
# mlab.figure(bgcolor=(1,1,1))
#
# s, mesh = plot_element()
# points = np.array([[0.01, 1, 1],
#                   [0.01, 1, -1],
#                   [0.01, -1, -1],
#                   [0.01, -1, 1]])*2
# tris=np.array([[0,1,2], [2,3,0]])
# mesh2 = trimesh.Trimesh(points, tris)
# for ii in range(7):
#    mesh2 =mesh2.subdivide()
#
# U = compute_U(mesh, mesh2.vertices) @ scalars
#
# s3= mlab.triangular_mesh(*mesh2.vertices.T, mesh2.faces, scalars=U, colormap='bwr')
# s3.enable_contours = True
# s3.contour.number_of_contours = 31
# s3.contour.filled_contours = True
# s3.contour.minimum_contour = -5.38e-07
# s3.contour.maximum_contour = 5.38e-07
##s3.actor.property.render_lines_as_tubes = True
# s3.actor.property.line_width = 3.0

#%%
# mlab.figure(bgcolor=(1,1,1))

# s, mesh = plot_element()
# points = np.array([[1, 1, -0.01],
#                   [1, -1, -0.01],
#                   [-1, -1, -0.01],
#                   [-1, 1, -0.01]])*2
# points[:,2] += 0.2
# tris=np.array([[0,1,2], [2,3,0]])
# mesh3 = trimesh.Trimesh(points, tris)
# for ii in range(3):
#    mesh3 =mesh3.subdivide()
# A = compute_A(mesh, mesh3.vertices) @ scalars
# vectors = mlab.quiver3d(*mesh3.vertices.T, *A, mode='arrow', color=(0,0,1))
# vectors.glyph.glyph_source.glyph_position = 'center'
#%%
# mlab.figure(bgcolor=(1,1,1))
# s, mesh = plot_element()
# points = np.array([[0.001, 1, 1],
#                   [0.001, 1, -1],
#                   [0.001, -1, -1],
#                   [0.001, -1, 1]])*2 + 0.001
p1 = np.array([-0.1, -1, 0]) * 2 + 0.0001
p2 = np.array([0.1, 1, 1]) * 2 + 0.0001
X, Y, Z = np.meshgrid(*np.linspace(p1, p2, 50).T, indexing="ij")
R = np.array([X.flatten(), Y.flatten(), Z.flatten()])
# B0 = compute_C(mesh, R.T) @ scalars
B1 = compute_C_analytic(mesh, R.T) @ scalars
B1[0] = 0
vecfield = mlab.pipeline.vector_field(X, Y, Z, *B1.T.reshape(3, 50, 50, 50))
streamline = mlab.pipeline.streamline(vecfield)
streamline.seed.widget = streamline.seed.widget_list[1]
# streamline.seed.widget = <tvtk.tvtk_classes.line_widget.LineWidget object at 0x7f96be04f678>
streamline.stream_tracer.start_position = np.array([0.0, 0.0, 0.0])
streamline.stream_tracer.integration_direction = "both"
streamline.seed.widget.resolution = 15
streamline.seed.widget.point1 = np.array([0.01, -0.6, 0.055])
streamline.seed.widget.point2 = np.array([0.01, 0.6, 0.05523521])
streamline.actor.property.render_lines_as_tubes = True
streamline.actor.property.line_width = 3.0
streamline.actor.property.color = (0, 0, 0)
streamline.update_streamlines = 0

# The other half
p1 = np.array([-0.1, -1, -1]) * 2 - 0.0001
p2 = np.array([0.1, 1, 0]) * 2 - 0.0001
X, Y, Z = np.meshgrid(*np.linspace(p1, p2, 50).T, indexing="ij")
R = np.array([X.flatten(), Y.flatten(), Z.flatten()])
# B0 = compute_C(mesh, R.T) @ scalars
B1 = compute_C_analytic(mesh, R.T) @ scalars
B1[0] = 0
vecfield = mlab.pipeline.vector_field(X, Y, Z, *B1.T.reshape(3, 50, 50, 50))
streamline = mlab.pipeline.streamline(vecfield)
streamline.seed.widget = streamline.seed.widget_list[1]
# streamline.seed.widget = <tvtk.tvtk_classes.line_widget.LineWidget object at 0x7f96be04f678>
streamline.stream_tracer.start_position = np.array([0.0, 0.0, 0.0])
streamline.stream_tracer.integration_direction = "both"
streamline.seed.widget.resolution = 15
streamline.seed.widget.point1 = np.array([0.01, -0.6, -0.055])
streamline.seed.widget.point2 = np.array([0.01, 0.6, -0.05523521])
streamline.actor.property.render_lines_as_tubes = True
streamline.actor.property.line_width = 3.0
streamline.actor.property.color = (0, 0, 0)
streamline.update_streamlines = 0


# vectors = mlab.quiver3d(*R, *B1.T, mode='arrow', color=(1,0,1))
# vectors.glyph.glyph_source.glyph_position = 'center'
# vectors.actor.property.render_lines_as_tubes = True
# vectors.actor.property.line_width = 3.0



from bfieldtools import mesh_calculus
from bfieldtools.utils import load_example_mesh

import pytest

import numpy as np
from numpy.testing import assert_allclose


def test_laplacian():

    mesh = load_example_mesh("unit_disc")

    L = mesh_calculus.laplacian_matrix(mesh)


def test_mass():

    mesh = load_example_mesh("unit_disc")

    M = mesh_calculus.mass_matrix(mesh, lumped=True)
    Ml = mesh_calculus.mass_matrix(mesh, lumped=False)


def test_others():

    mesh = load_example_mesh("unit_disc")

    vals = np.ones((len(mesh.vertices),)) - np.linalg.norm(mesh.vertices, axis=1)

    G = mesh_calculus.gradient(vals, mesh, rotated=False)
    Gr = mesh_calculus.gradient(vals, mesh, rotated=True)

    D = mesh_calculus.divergence(Gr.T, mesh)
    C = mesh_calculus.curl(Gr.T, mesh)



"""

Tools for calculating the surface harmonic representation of
the magnetic field. Surface harmonics can represent
of any stream function as a series.

Surface harmonics == Laplace-Beltrami eigenfunctions

"""

__all__ = ["SuhBasis"]

import trimesh
from scipy.sparse.linalg import eigsh
from scipy.linalg import eigh

import numpy as np

from .viz import plot_data_on_vertices
from . import mesh_conductor


class SuhBasis:
    """

    Class for representing magnetic field using surface harmonics

    """

    def __init__(
        self,
        obj,
        Nc=None,
        boundary_condition="dirichlet",
        magnetic=False,
        solver_sparse=True,
        **kwargs
    ):
        """
        Parameters
        ----------
        obj : Trimesh-object or Conductor-object
            Represents the boundary on which current density is specified
            or Conductor object that wraps the mesh
        Nc : Number of components
            If None (default), compute all components.
        boundary_condition : str  "dirichlet" (default) or "neumann"
            if zero-Dirichlet boundary conditions ("dirichlet")
            are used the basis corresponds to inner_vertices
            else with zero-Neumann condition ("neumann")
            the basis corresponds to all vertices
        magnetic: False or 'DC' or 'AC'
            Determines eigenvalue equation. If False, use laplacian and mass matrices.
            If 'DC', use resistance and mass matrices. If 'AC', use resistance and inductance matrices.
        solver_sparse: Boolean (True)
            If True, use solver from scipy.sparse.linalg rather than scipy.linalg
        kwargs: dict
            Passed to Conductor creation if a Trimesh object is passed as 'obj'

        """

        if boundary_condition in ("dirichlet", "neumann"):
            self.bc = boundary_condition
        else:
            raise ValueError(
                "boundary_conditions should be either dirichlet or neumann"
            )

        if isinstance(obj, mesh_conductor.MeshConductor):
            if obj.opts["resistance_full_rank"] is True:
                raise ValueError(
                    "resistance matrix must not be deflated (resistance_full_rank must be False)"
                )
            self.mesh_conductor = obj
        elif isinstance(obj, trimesh.Trimesh):
            self.mesh_conductor = mesh_conductor.MeshConductor(
                mesh_obj=obj, resistance_full_rank=False, **kwargs
            )
        else:
            raise TypeError("obj type should be either Trimesh or Conductor")
        self.mesh = self.mesh_conductor.mesh

        self.magnetic = magnetic
        self.solver_sparse = solver_sparse

        if self.bc == "neumann":
            self.mesh_conductor.set_basis("vertex")
        else:
            self.mesh_conductor.set_basis("inner")

        if self.mesh.is_watertight or self.bc == "neumann":
            self._max_Nc = self.mesh_conductor.basis.shape[1] - 1
        else:
            self._max_Nc = self.mesh_conductor.basis.shape[1]

        if Nc is None:
            print("No component count given, computing all components.")
            self.Nc = self._max_Nc
        elif Nc > self._max_Nc:
            raise ValueError("Mesh supports up to %d components" % self._max_Nc)
        else:
            self.Nc = Nc

        self.inner_vertices = self.mesh_conductor.inner_vertices
        self.holes = self.mesh_conductor.holes
        self.inner2vert = self.mesh_conductor.inner2vert

        self.calculate_basis()

    def calculate_basis(self, shiftinvert=True, v0=None):
        """ Calculate basis functions as eigenfunctions of matrices
        specified by the 'magnetic' parameter
        
        Parameters
        ----------
        shiftinvert: Boolean (True)
            use shiftinvert mode to calculate eigenstuff faster (experimental)
        v0: array (b, b)
            Intial value matrix for eigenvalue decomposition algorithm. Shape depends on BC: if neumann b=self.mesh_conductor.basis.shape[0],
            if dirichlet b=self.mesh_conductor.basis.shape[1]
                        
                    
        """
        print("Calculating surface harmonics expansion...")

        if not self.magnetic:
            L = -self.mesh_conductor.laplacian
            M = self.mesh_conductor.mass
        elif self.magnetic == "AC":
            L = self.mesh_conductor.resistance
            M = self.mesh_conductor.inductance
            # If mesh is closed, add 'deflation' to the inductance
            # L matrix deflation is already provided by Conductor(resistance_full_rank=True)
            if self.mesh.is_watertight:
                M += np.ones(M.shape) * np.mean(np.diag(M))
        elif self.magnetic == "DC":
            L = self.mesh_conductor.resistance
            M = self.mesh_conductor.mass
        else:
            raise ValueError("Parameter 'magnetic' must be False, 'DC', or 'AC'.")

        if self.mesh.is_watertight or self.bc == "neumann":
            print("Closed mesh or Neumann BC, leaving out the constant component")

            N0 = 1

            N = np.min((self.Nc + 1, self._max_Nc))

            # adjust self.Nc so that is corresponds to suh.basis.shape[1]
            if N == self._max_Nc:
                self.Nc -= 1
        else:
            N0 = 0
            N = self.Nc

        # Make sure that matrices are symmetric, positive definite
        L = 0.5 * (L + L.T)
        M = 0.5 * (M + M.T)

        if v0 is None:
            v0 = np.ones(L.shape[1])  # avoid random basis for symmetric geometries

        # Use sparse solver if possible
        # 1. matrices have to be sparse (i.e. magnetic=False)
        # 2. solver_sparse = True
        # 3. the sparse solver cannot find all eigenvalues, so Nc < max_Nc
        if (not self.magnetic) and self.solver_sparse and (self.Nc < self._max_Nc):

            if shiftinvert:
                u, v = eigsh(L, N, M, sigma=0, which="LM", v0=v0)
            else:
                u, v = eigsh(L, N, M, which="SA", v0=v0)

            self.basis = v[:, N0:]
            self.eigenvals = u[N0:]

        else:
            if not self.magnetic:
                L = L.toarray()
                M = M.toarray()

                u, v = eigh(L, M, eigvals=(0, N - 1))
            else:
                if self.magnetic == "DC":
                    M = M.toarray()

                u, v = eigh(L, M, eigvals=(0, N - 1))

            self.basis = v[:, N0:]
            self.eigenvals = u[N0:]

    def field(self, coeffs, points):
        """ Calculate field at points

            Parameters

            coeffs : (self.Nc,) array of basis function coefficients
            points : (N_points, 3) field evaluation points

            Returns:

                field : (N_points, 3) magnetic field
        """
        return self.basis_fields(points) @ coeffs

    def basis_fields(self, points):
        """ Calculate basis fields at points

            Return:

                Fields (3, N_points, self.Nc)
        """
        B_coupling = self.mesh_conductor.B_coupling(points)

        return B_coupling @ self.basis

    def fit_coeffs(self, points, data):
        """ Fit basis function coefficients to the data
        """
        assert len(data) > self.Nc

        A = self.basis_fields(points).reshape(-1, self.Nc)
        b = data.T.flatten()
        x, res, rank, s = np.linalg.lstsq(A, b, rcond=None)

        if rank < self.Nc:
            print("Matrix rank not full, result might be inaccurate")
        return x

    def plot(
        self, Nfuncs, dist=0.5, Ncols=None, figure=None, figsize=(800, 800), **kwargs
    ):
        """ Plot basis functions on the mesh

            Nfuncs: int or array-like
                 if int, the number functions starting from the first'Blm',
                 if list/array: the indices of the functions

            dist: float
                distance between the plotted objects relative to their size

            Ncols: int or None
                the number of columns in the plot
                If none automatically determined

            fig: handle for mlab figure

            figsize: size of a new figure if 'fig' not given

            ncolors:
                number of colors in the colormap

            kwargs: keyword arguments passed to mayavi (colormap, etc.)

        """
        from mayavi import mlab

        if figure is None:
            figure = mlab.figure(
                None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=figsize
            )

        if isinstance(Nfuncs, int):
            N = Nfuncs
            indices = np.arange(Nfuncs)
        else:
            indices = Nfuncs
            N = len(indices)

        if Ncols is None:
            Ncols = np.floor(np.sqrt(N) + 1)

        dx = (self.mesh.vertices[:, 0].max() - self.mesh.vertices[:, 0].min()) * (
            1 + dist
        )
        dy = (self.mesh.vertices[:, 1].max() - self.mesh.vertices[:, 1].min()) * (
            1 + dist
        )

        i = 0
        j = 0

        s = []
        for n in indices:
            print(i, j)

            tmp_mesh = self.mesh.copy()
            tmp_mesh.vertices[:, 0] += i * dx
            tmp_mesh.vertices[:, 1] -= j * dy

            if self.bc == "neumann":
                scalars = self.basis[:, n]
            else:
                scalars = self.inner2vert @ self.basis[:, n]

            s.append(plot_data_on_vertices(tmp_mesh, scalars, figure=figure, **kwargs))

            if i < Ncols:
                i += 1
            else:
                j += 1
                i = 0

        return s



"""
Line charge
================

Test and validation of gamma_0

For the math see:
        A. S. Ferguson, Xu Zhang and G. Stroink,
        "A complete linear discretization for calculating the magnetic field
        using the boundary element method,"
        in IEEE Transactions on Biomedical Engineering,
        vol. 41, no. 5, pp. 455-460, May 1994.
        doi: 10.1109/10.293220

"""

import numpy as np
import matplotlib.pyplot as plt
import sys
from mayavi import mlab

path = "/m/home/home8/80/makinea1/unix/pythonstuff/bfieldtools"
if path not in sys.path:
    sys.path.insert(0, path)

from bfieldtools.integrals import gamma0
from bfieldtools.utils import tri_normals_and_areas

#%%
#%% Test potential shape slightly above the surface
points = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])

tris = np.array([[0, 1, 2]])
p_tris = points[tris]

# Evaluation points
Nx = 100
xx = np.linspace(-2, 2, Nx)
X, Y = np.meshgrid(xx, xx, indexing="ij")
Z = np.zeros_like(X) + 0.1
p_eval = np.array([X, Y, Z]).reshape(3, -1).T

# Difference vectors
RR = p_eval[:, None, None, :] - p_tris[None, :, :, :]

pot = gamma0(RR)

# Plot shape
plt.figure()
plt.imshow(
    pot[:, 0].sum(axis=-1).reshape(Nx, Nx),
    extent=(xx.min(), xx.max(), xx.max(), xx.min()),
)
plt.ylabel("x")
plt.xlabel("y")

#%%
""" Test potential at directly at the edge. As the line has no
    perpendicular dimensions, the potential is infinite at the edge.
    The regularization factor given to the function apprximates the line
    current with a small radius, giving rougly constant potential on the line.
    The relative error between the regularized verison and
    the infinitely thin line charge seems to be on the order of the "reg" value

    The "symmetrize" option symmeterizes the result with respect to the
    mid point. This removes errors (Nans) on the other continuoation of the edge
"""
#%%
points = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])

tris = np.array([[0, 1, 2]])
p_tris = points[tris]

# Evaluation points
Nx = 1000
x = np.linspace(-2, 2, Nx)
y = z = np.zeros_like(x)
p_eval = np.array([x, y, z]).T

# Difference vectors
RR = p_eval[:, None, None, :] - p_tris[None, :, :, :]

# Regularize and symmetrize
pot = gamma0(RR, 1e-13, True)
pot0 = pot[:, 0, 2]
plt.figure()
plt.plot(x, pot0, linewidth=5)
# Symmetrize, but do not regularize
pot = gamma0(RR, 0, True)
pot1 = pot[:, 0, 2]
plt.plot(x, pot1, "--", linewidth=3)
# Neither
pot = gamma0(RR, 0, False)
pot2 = pot[:, 0, 2]
plt.plot(x, pot2)
plt.xlabel("x")
plt.legend(("reg + sym", "sym", "neither"))

plt.figure()
plt.title("Relative error")
plt.plot(x, abs(pot0 - pot1) / pot1)



"""
Interpolate stream function
===========================

Minimal example showing how to subdivide a mesh and interpolate a scalar function
defined on that mesh to match.

"""

from tvtk.api import tvtk
from mayavi import mlab
import trimesh

import numpy as np
from scipy.linalg import eigh

from bfieldtools.mesh_calculus import laplacian_matrix, mass_matrix
from bfieldtools import utils

import pkg_resources


#%%
# Load a simple mesh and compute an example scalar function on it.
# In this case, the scalar function is an eigenvector of a generalized eigenvalue decomposition

mesh = trimesh.load(
    pkg_resources.resource_filename("bfieldtools", "example_meshes/10x10_plane.obj")
)

boundaries, inner_verts = utils.find_mesh_boundaries(mesh)

L = laplacian_matrix(mesh)
M = mass_matrix(mesh)

u, v = eigh(
    -L.todense()[inner_verts][:, inner_verts], M.todense()[inner_verts][:, inner_verts]
)

scalars = np.zeros(mesh.vertices.shape[0])
scalars[inner_verts] = v[:, 12]


original_scalars = scalars.copy()
original_mesh = mesh.copy()

#%%
# Plot original scalars and mesh

scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))

mlab.triangular_mesh(
    *original_mesh.vertices.T,
    original_mesh.faces,
    scalars=original_scalars,
    representation="wireframe"
)


#%%
# Now, interpolate scalars


ug = tvtk.UnstructuredGrid(points=mesh.vertices)

ug.set_cells(tvtk.Triangle().cell_type, mesh.faces)
ug.point_data.scalars = scalars
ug.point_data.scalars.name = "scalars"


mesh = original_mesh.subdivide().subdivide()
scalars = mlab.pipeline.probe_data(ug, *mesh.vertices.T)


#%%
# Plot subdivided mesh and interpolated scalars

scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))

mlab.triangular_mesh(
    *mesh.vertices.T, mesh.faces, scalars=scalars, representation="wireframe"
)



from bfieldtools import utils

import numpy as np

import pytest

from .test_line_conductor import _fake_line_conductor


def test_mesh_utils():

    mesh1 = utils.load_example_mesh("unit_disc")
    mesh2 = utils.load_example_mesh("10x10_plane")

    mesh2.vertices += np.array([[0, 0, 20]])

    mesh3 = utils.combine_meshes((mesh1, mesh2))

    assert len(mesh3.vertices) == len(mesh1.vertices) + len(mesh2.vertices)


@pytest.mark.xfail
def test_load_mesh_fail():
    mesh = utils.load_example_mesh("this mesh does not exist")


def test_quad_points():
    mesh = utils.load_example_mesh("unit_disc")

    w, qp = utils.get_quad_points(mesh.vertices, mesh.faces)
    w, qp = utils.get_quad_points(mesh.vertices, mesh.faces, "dunavant_02")
    w, qp = utils.get_quad_points(mesh.vertices, mesh.faces, method="lether", index=3)


@pytest.mark.xfail
def test_quad_points_fail1():
    mesh = utils.load_example_mesh("unit_disc")
    w, qp = utils.get_quad_points(mesh.vertices, mesh.faces, method="lether")


@pytest.mark.xfail
def test_quad_points_fail2():
    mesh = utils.load_example_mesh("unit_disc")
    w, qp = utils.get_quad_points(mesh.vertices, mesh.faces, method="dunavant")


def test_line_quad_points():
    lp = _fake_line_conductor()

    w, qp = utils.get_line_quad_points(lp.vertices[lp.entities[0].points])
    w, qp = utils.get_line_quad_points(lp.vertices[lp.entities[0].points], "midpoint")
    w, qp = utils.get_line_quad_points(lp.vertices[lp.entities[0].points], "fejer_1", 1)


@pytest.mark.xfail
def test_line_quad_points_fail1():
    lp = _fake_line_conductor()

    w, qp = utils.get_line_quad_points(lp.vertices[lp.entities[0].points], "fejer_1")


@pytest.mark.xfail
def test_line_quad_points_fail2():
    lp = _fake_line_conductor()

    w, qp = utils.get_line_quad_points(
        lp.vertices[lp.entities[0].points], method="dunavant"
    )


def test_cylinder_points():

    cp = utils.cylinder_points()
    cp = utils.cylinder_points(alpha=90)
    cp = utils.cylinder_points(orientation=np.array([0, 1, 0]))



"""
Coil with minimal eddy currents
===============================
Compact example of design of a cylindrical coil surrounded by a RF shield, i.e. a conductive surface.
The effects of eddy currents due to inductive interaction with the shield is minimized
"""

import numpy as np
from mayavi import mlab
import trimesh


from bfieldtools.mesh_conductor import MeshConductor

from bfieldtools.coil_optimize import optimize_streamfunctions
from bfieldtools.contour import scalar_contour
from bfieldtools.viz import plot_3d_current_loops, plot_data_on_vertices

import pkg_resources

from pyface.api import GUI

_gui = GUI()


# Set unit, e.g. meter or millimeter.
# This doesn't matter, the problem is scale-invariant
scaling_factor = 1


# Load example coil mesh that is centered on the origin
coilmesh = trimesh.load(
    file_obj=pkg_resources.resource_filename(
        "bfieldtools", "example_meshes/open_cylinder.stl"
    ),
    process=True,
)

angle = np.pi / 2
rotation_matrix = np.array(
    [
        [np.cos(angle), 0, np.sin(angle), 0],
        [0, 1, 0, 0],
        [-np.sin(angle), 0, np.cos(angle), 0],
        [0, 0, 0, 1],
    ]
)

coilmesh.apply_transform(rotation_matrix)

coilmesh1 = coilmesh.copy()

coilmesh2 = coilmesh.copy()


# Create mesh class object
coil = MeshConductor(
    verts=coilmesh1.vertices * 0.75,
    tris=coilmesh1.faces,
    fix_normals=True,
    basis_name="suh",
    N_suh=400,
)


def alu_sigma(T):
    ref_T = 293  # K
    ref_rho = 2.82e-8  # ohm*meter
    alpha = 0.0039  # 1/K

    rho = alpha * (T - ref_T) * ref_rho + ref_rho

    return 1 / rho


resistivity = 1 / alu_sigma(T=293)  # room-temp Aluminium
thickness = 0.5e-3  # 0.5 mm thick


# Separate object for shield geometry
shield = MeshConductor(
    verts=coilmesh2.vertices.copy() * 1.1,
    tris=coilmesh2.faces.copy(),
    fix_normals=True,
    basis_name="inner",
    resistivity=resistivity,
    thickness=thickness,
)


#%%
# Set up target  points and plot geometry


center = np.array([0, 0, 0])

sidelength = 0.25 * scaling_factor
n = 12
xx = np.linspace(-sidelength / 2, sidelength / 2, n)
yy = np.linspace(-sidelength / 2, sidelength / 2, n)
zz = np.linspace(-sidelength / 2, sidelength / 2, n)
X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

x = X.ravel()
y = Y.ravel()
z = Z.ravel()

target_points = np.array([x, y, z]).T

# Turn cube into sphere by rejecting points "in the corners"
target_points = (
    target_points[np.linalg.norm(target_points, axis=1) < sidelength / 2] + center
)


# Plot coil, shield and target points
f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
coil.plot_mesh(figure=f, opacity=0.2)
shield.plot_mesh(figure=f, opacity=0.2)
mlab.points3d(*target_points.T)


#%%
# Compute eddy-current coupling

mutual_inductance = coil.mutual_inductance(shield)

# Take into account the field produced by currents induced into the shield
# NB! This expression is for instantaneous step-function switching of coil current, see Eq. 18 in G.N. Peeren, 2003.

shield.M_coupling = np.linalg.solve(-shield.inductance, mutual_inductance.T)
secondary_C = shield.B_coupling(target_points) @ -shield.M_coupling

#%%
# Create bfield specifications used when optimizing the coil geometry

# The absolute target field amplitude is not of importance,
# and it is scaled to match the C matrix in the optimization function

target_field = np.zeros(target_points.shape)
target_field[:, 1] = target_field[:, 1] + 1


target_spec = {
    "coupling": coil.B_coupling(target_points),
    "abs_error": 0.01,
    "target": target_field,
}


from scipy.linalg import eigh

l, U = eigh(shield.resistance, shield.inductance, eigvals=(0, 500))


time = [0.001, 0.003, 0.005]
eddy_error = [0.05, 0.01, 0.0025]
# time_decay = U @ np.exp(-l[None, :]*time[:, None]) @ np.pinv(U)

time_decay = np.zeros(
    (len(time), shield.inductance.shape[0], shield.inductance.shape[1])
)

induction_spec = []


Uinv = np.linalg.pinv(U)
for idx, t in enumerate(time):
    time_decay = U @ np.diag(np.exp(-l * t)) @ Uinv
    eddy_coupling = shield.B_coupling(target_points) @ time_decay @ shield.M_coupling
    induction_spec.append(
        {
            "coupling": eddy_coupling,
            "abs_error": eddy_error[idx],
            "rel_error": 0,
            "target": np.zeros_like(target_field),
        }
    )

#%%
# Run QP solver to optimize stream function

import mosek

coil.s, prob = optimize_streamfunctions(
    coil,
    [target_spec] + induction_spec,
    objective="minimum_inductive_energy",
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)

from bfieldtools.mesh_conductor import StreamFunction

shield.induced_s = StreamFunction(shield.M_coupling @ coil.s, shield)

#%%
# Plot coil windings and target points


loops = scalar_contour(coil.mesh, coil.s.vert, N_contours=6)


f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(600, 500))
mlab.clf()

plot_3d_current_loops(loops, colors="auto", figure=f, tube_radius=0.005)

B_target = coil.B_coupling(target_points) @ coil.s

mlab.quiver3d(*target_points.T, *B_target.T)

shield.plot_mesh(
    representation="surface",
    opacity=0.5,
    cull_back=True,
    color=(0.8, 0.8, 0.8),
    figure=f,
)
shield.plot_mesh(
    representation="surface",
    opacity=1,
    cull_front=True,
    color=(0.8, 0.8, 0.8),
    figure=f,
)

f.scene.camera.parallel_projection = 1

f.scene.camera.zoom(1.4)

#%%
# For comparison, let's see how the coils look when we ignore the conducting shield


coil.unshielded_s, coil.unshielded_prob = optimize_streamfunctions(
    coil,
    [target_spec],
    objective="minimum_inductive_energy",
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)

shield.unshielded_induced_s = StreamFunction(
    shield.M_coupling @ coil.unshielded_s, shield
)

loops = scalar_contour(coil.mesh, coil.unshielded_s.vert, N_contours=6)

f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(600, 500))
mlab.clf()

plot_3d_current_loops(loops, colors="auto", figure=f, tube_radius=0.005)

B_target_unshielded = coil.B_coupling(target_points) @ coil.unshielded_s

mlab.quiver3d(*target_points.T, *B_target_unshielded.T)

shield.plot_mesh(
    representation="surface",
    opacity=0.5,
    cull_back=True,
    color=(0.8, 0.8, 0.8),
    figure=f,
)
shield.plot_mesh(
    representation="surface",
    opacity=1,
    cull_front=True,
    color=(0.8, 0.8, 0.8),
    figure=f,
)

f.scene.camera.parallel_projection = 1

f.scene.camera.zoom(1.4)


#%%
# Finally, let's compare the time-courses


tmin, tmax = 0, 0.025
Fs = 2000

time = np.linspace(tmin, tmax, int(Fs * (tmax - tmin) + 1))

time_decay = np.zeros(
    (len(time), shield.inductance.shape[0], shield.inductance.shape[1])
)

Uinv = np.linalg.pinv(U)
for idx, t in enumerate(time):
    time_decay[idx] = U @ np.diag(np.exp(-l * t)) @ Uinv


B_t = shield.B_coupling(target_points) @ (time_decay @ shield.induced_s).T

unshieldedB_t = (
    shield.B_coupling(target_points) @ (time_decay @ shield.unshielded_induced_s).T
)

import matplotlib.pyplot as plt


fig, ax = plt.subplots(1, 1, sharex=True, figsize=(8, 4))
ax.plot(
    time * 1e3,
    np.mean(np.linalg.norm(B_t, axis=1), axis=0).T,
    "k-",
    label="Minimized",
    linewidth=1.5,
)
ax.set_ylabel("Transient field amplitude")
ax.semilogy(
    time * 1e3,
    np.mean(np.linalg.norm(unshieldedB_t, axis=1), axis=0).T,
    "k--",
    label="Ignored",
    linewidth=1.5,
)
ax.set_xlabel("Time (ms)")


ax.set_ylim(1e-4, 0.5)
ax.set_xlim(0, 25)


plt.grid(which="both", axis="y", alpha=0.1)

plt.legend()
fig.tight_layout()

ax.vlines([1, 5, 10, 20], 1e-4, 0.5, alpha=0.1, linewidth=3, color="r")



# -*- coding: utf-8 -*-
"""
Validation of self inductance using spherical harmonics
====================================================================

Study the eigenvalue spectrum of the discretize self-inductance operator
on a spherical mesh. Compare the spectrum to analytical solution
"""

import numpy as np
from bfieldtools.mesh_impedance import self_inductance_matrix, mutual_inductance_matrix
from bfieldtools.mesh_calculus import mass_matrix
from bfieldtools.utils import load_example_mesh
from scipy.linalg import eigh
import matplotlib.pyplot as plt
import trimesh

# This is icosphere(4)?
# mesh = load_example_mesh("unit_sphere")
# The test is faster with a smaller number of vertices
mesh = trimesh.creation.icosphere(3)

Nvals = 150
uu1 = []
uu2 = []
vv1 = []
vv2 = []
M = mass_matrix(mesh)

for q in range(4):
    L = self_inductance_matrix(mesh, analytic_self_coupling=True, quad_degree=q + 1)
    uu, vv = eigh(L, M.toarray(), eigvals=(0, Nvals - 1))
    uu1.append(uu)
    vv1.append(vv)

    # L = self_inductance_matrix(mesh, analytic_self_coupling=False,
    # quad_degree=q+1)
    L = mutual_inductance_matrix(mesh, mesh, quad_degree=q + 1)
    uu, vv = eigh(L, M.toarray(), eigvals=(0, Nvals - 1))
    uu2.append(uu)
    vv2.append(vv)


#%%
""" 
 Spherical harmonics are the eigenfunctions of self-inductance operator
 The correct eigenvalues derived using Taulu 2005 Eqs. (22, 23, A1, A5, A6)
 By considering the normal component of the magnetic field produced by 
 a single Y_lm. The result is
 e = mu_0*(l*(l+1)/(2*l+1))/R
"""
R = np.linalg.norm(mesh.vertices[mesh.faces].mean(axis=1), axis=-1).mean()
mu0 = (1e-7) * 4 * np.pi
ll = np.array([l for l in range(20) for m in range(-l, l + 1)])
ll = ll[:Nvals]
evals = ll * (ll + 1) / (2 * ll + 1)


#%%
# plt.plot(evals, 'k')
for u in uu1:
    uu_scaled = u / mu0 * R
    plt.plot(abs(evals[1:] - uu_scaled[1:]) / evals[1:], "-")
    # plt.plot(uu_scaled)
plt.legend(("qd=1", "qd=2", "qd=3", "qd=4"))
plt.gca().set_prop_cycle(None)
for u in uu2:
    uu_scaled = u / mu0 * R
    plt.plot(abs(evals[1:] - uu_scaled[1:]) / evals[1:], "--")
plt.title("solid using self_inductance, dashed using mutual_inductance")
plt.xlabel("# eigenvalue")
plt.ylabel("Relative error")

#%% Eigenfunctions
from bfieldtools.utils import MeshProjection
from bfieldtools.sphtools import ylm
from bfieldtools.sphtools import cartesian2spherical

ylm_on_hats = []
i1 = 0
vv1_projs = np.zeros((len(vv1), vv1[0].shape[1]))
vv2_projs = np.zeros((len(vv2), vv2[0].shape[1]))
mp = MeshProjection(mesh, 4)


for l in range(0, 13):
    i0 = i1
    print(f"l={l}")
    for m in range(-l, l + 1):

        def func(r):
            sphcoords = cartesian2spherical(r)
            return ylm(l, m, sphcoords[:, 1], sphcoords[:, 2])

        ylm_on_hats.append(mp.hatfunc_innerproducts(func))
        i1 += 1
    for ii, vv in enumerate(vv1):
        # Project self-inductance eigenfunctions to l-subspace
        p = np.sum((np.array(ylm_on_hats[i0:i1]) @ vv[:, i0:i1]) ** 2, axis=0)
        vv1_projs[ii, i0:i1] = p
    for ii, vv in enumerate(vv2):
        # Project self-inductance eigenfunctions to l-subspace
        p = np.sum((np.array(ylm_on_hats[i0:i1]) @ vv[:, i0:i1]) ** 2, axis=0)
        vv2_projs[ii, i0:i1] = p

#%%
plt.figure()
eff_R2 = 1  # np.sum(mesh.area_faces) / (4 * np.pi)
plt.plot(vv1_projs.T / eff_R2, "-")
plt.legend(("qd=1", "qd=2", "qd=3", "qd=4"))
plt.gca().set_prop_cycle(None)
plt.plot(vv2_projs.T / eff_R2, "--")
plt.title("solid using self_inductance, dashed using mutual_inductance")

plt.xlabel("# eigenfunction")
plt.ylabel("Squared norm in L-subspace")



"""
SUH-SPH interpolation comparison
==================================
"""


import numpy as np
from bfieldtools.mesh_conductor import MeshConductor, StreamFunction
from mayavi import mlab
import trimesh
import matplotlib.pyplot as plt

from bfieldtools.sphtools import basis_fields as sphfield
from bfieldtools.sphtools import field as sph_field_eval
from bfieldtools.sphtools import basis_potentials, potential
import mne

from bfieldtools.viz import plot_data_on_vertices, plot_mesh

#%%
SAVE_DIR = "./MNE interpolation/"

#%%
EVOKED = True

with np.load(SAVE_DIR + "mne_data.npz", allow_pickle=True) as data:
    p = data["p"]
    n = data["n"]
    mesh = trimesh.Trimesh(vertices=data["vertices"], faces=data["faces"])

if EVOKED:
    evoked = mne.Evoked(SAVE_DIR + "left_auditory-ave.fif")

    i0, i1 = evoked.time_as_index(0.08)[0], evoked.time_as_index(0.09)[0]
    field = evoked.data[:, i0:i1].mean(axis=1)

else:
    # take "data" from lead field matrix, i.e, topography of a single dipole
    from mne.datasets import sample
    import os

    data_path = sample.data_path()

    raw_fname = data_path + "/MEG/sample/sample_audvis_raw.fif"
    trans = data_path + "/MEG/sample/sample_audvis_raw-trans.fif"
    src = data_path + "/subjects/sample/bem/sample-oct-6-src.fif"
    bem = data_path + "/subjects/sample/bem/sample-5120-5120-5120-bem-sol.fif"
    subjects_dir = os.path.join(data_path, "subjects")

    # Note that forward solutions can also be read with read_forward_solution
    fwd = mne.make_forward_solution(
        raw_fname, trans, src, bem, meg=True, eeg=False, mindist=5.0, n_jobs=2
    )
    # Take only magnetometers
    mags = np.array([n[-1] == "1" for n in fwd["sol"]["row_names"]])
    L = fwd["sol"]["data"][mags, :]
    # Take the first dipole
    field = L[:, 56]

#%% radius for inner/outer sph

R = np.min(np.linalg.norm(p, axis=1)) - 0.02

#%%

lmax = 7  # maximum degree
Bca, Bcb = sphfield(p, lmax, normalization="energy", R=R)

# sph-components at sensors
Bca_sensors = np.einsum("ijk,ij->ik", Bca, n)
Bcb_sensors = np.einsum("ijk,ij->ik", Bcb, n)


#%% Visualize sph components at the helmet
# idx = 20

# evoked1 = evoked.copy()
# evoked1.data[:, :] = np.tile(Bca_sensors[:, idx].T, (evoked.times.shape[0], 1)).T
# evoked1.plot_topomap(times=0.080, ch_type="mag", colorbar=False)

# evoked1 = evoked.copy()
# evoked1.data[:, :] = np.tile(Bcb_sensors[:, idx].T, (evoked.times.shape[0], 1)).T
# evoked1.plot_topomap(times=0.080, ch_type="mag", colorbar=False)

#%% calculate inner sph-coeffients with pinv
PINV = True
if PINV:
    alpha = np.linalg.pinv(Bca_sensors, rcond=1e-15) @ field
else:
    # Calculate using regularization
    ssa = np.linalg.svd(Bca_sensors @ Bca_sensors.T, False, False)
    reg_exp = 6
    _lambda = np.max(ssa) * (10 ** (-reg_exp))
    # angular-Laplacian in the sph basis is diagonal
    La = np.diag([l * (l + 1) for l in range(1, lmax + 1) for m in range(-l, l + 1)])
    BB = Bca_sensors.T @ Bca_sensors + _lambda * La
    alpha = np.linalg.solve(BB, Bca_sensors.T @ field)

# Reconstruct field in helmet

# reco_sph = np.zeros(field.shape)
# i = 0
# for l in range(1, lmax + 1):
#     for m in range(-1 * l, l + 1):
#         reco_sph += alpha[i] * Bca_sensors[:, i]
#         i += 1

# Produces the same result as the loop
reco_sph = Bca_sensors @ alpha

print(
    "SPH-reconstruction relative error:",
    np.linalg.norm(reco_sph - field) / np.linalg.norm(field),
)

#%%
##%% Fit the surface current for the auditory evoked response using pinv
# c = MeshConductor(mesh_obj=mesh, basis_name="suh", N_suh=35)
# M = c.mass
# B_sensors = np.einsum("ijk,ij->ik", c.B_coupling(p), n)
#
#
# asuh = np.linalg.pinv(B_sensors, rcond=1e-15) @ field
#
# s = StreamFunction(asuh, c)
# b_filt = B_sensors @ s


#%% Suh fit

c = MeshConductor(mesh_obj=mesh, basis_name="suh", N_suh=150)
M = c.mass

B_sensors = np.einsum("ijk,ij->ik", c.B_coupling(p), n)
ss = np.linalg.svd(B_sensors @ B_sensors.T, False, False)

reg_exp = 1
plot_this = True
rel_errors = []
_lambda = np.max(ss) * (10 ** (-reg_exp))
# Laplacian in the suh basis is diagonal
BB = B_sensors.T @ B_sensors + _lambda * (-c.laplacian) / np.max(abs(c.laplacian))
a = np.linalg.solve(BB, B_sensors.T @ field)

s = StreamFunction(a, c)

reco_suh = B_sensors @ s

print(
    "SUH-reconstruction relative error:",
    np.linalg.norm(reco_suh - field) / np.linalg.norm(field),
)

f = mlab.figure(bgcolor=(1, 1, 1))
surf = s.plot(False, figure=f)
surf.actor.mapper.interpolate_scalars_before_mapping = True
surf.module_manager.scalar_lut_manager.number_of_colors = 16

#%% Plot the evoked and the reconsctructions
# evoked1 = evoked.copy()
# evoked1.data[:, :] = np.tile(field.T, (evoked.times.shape[0], 1)).T
# evoked1.plot_topomap(times=0.080, ch_type="mag")

# evoked1 = evoked.copy()
# evoked1.data[:, :] = np.tile(reco_sph.T, (evoked.times.shape[0], 1)).T
# evoked1.plot_topomap(times=0.080, ch_type="mag")


# evoked1 = evoked.copy()
# evoked1.data[:, :] = np.tile(reco_suh.T, (evoked.times.shape[0], 1)).T
# evoked1.plot_topomap(times=0.080, ch_type="mag")

#%% Plot spectra
fig, ax = plt.subplots(1, 1)
ax.plot(alpha ** 2)


L = np.zeros((0,))
M = np.zeros((0,))


for l in range(1, lmax + 1):
    m_l = np.arange(-l, l + 1, step=1, dtype=np.int_)
    M = np.append(M, m_l)
    L = np.append(L, np.repeat(l, len(m_l)))

xticknames = [None] * len(alpha)
for i in range(len(alpha)):
    xticknames[i] = str(M[i])

    m_l = np.arange(-L[i], L[i] + 1, step=1)

    if i == int(np.floor(len(m_l))):
        xticknames[i] += "\n" + str(L[i])


plt.figure()
plt.plot(a ** 2)


#%% Compute potential on the helmet mesh
from bfieldtools.utils import load_example_mesh
from bfieldtools.flatten_mesh import flatten_mesh, mesh2plane

helmet = load_example_mesh("meg_helmet", process=False)
# Bring the surface roughly to the correct place
helmet.vertices[:, 2] -= 0.045
# The helmet is slightly tilted, correct for this
# (probably the right coordinate transformation could be found from MNE)
rotmat = np.eye(3)
tt = 0.015 * np.pi
rotmat[:2, :2] = np.array([[np.cos(tt), np.sin(tt)], [-np.sin(tt), np.cos(tt)]])
helmet.vertices = helmet.vertices @ rotmat
tt = -0.02 * np.pi
rotmat[1:, 1:] = np.array([[np.cos(tt), np.sin(tt)], [-np.sin(tt), np.cos(tt)]])
helmet.vertices = helmet.vertices @ rotmat
helmet.vertices[:, 1] += 0.005

# plot_mesh(helmet)
# mlab.points3d(*p.T, scale_factor=0.01)


B_sph_helmet = sph_field_eval(
    helmet.vertices,
    alpha,
    np.zeros(alpha.shape),
    lmax=lmax,
    normalization="energy",
    R=R,
)
B_sph_helmet = np.einsum("ij,ij->i", B_sph_helmet, helmet.vertex_normals)
B_suh_helmet = c.B_coupling(helmet.vertices) @ s
B_suh_helmet = np.einsum("ij,ij->i", B_suh_helmet, helmet.vertex_normals)

#%% Compute flattened mesh


u, v, helmet2d = flatten_mesh(helmet, 0.9)
puv = mesh2plane(p, helmet, u, v)


#%% Magnetic field at sensor array surface

from scipy.interpolate import Rbf

rbf_f = Rbf(puv[:, 0], puv[:, 1], field, function="linear", smooth=0)
rbf_field = rbf_f(helmet2d.vertices[:, 0], helmet2d.vertices[:, 1])


vmin = -7e-13
vmax = 7e-13
f = plot_data_on_vertices(helmet2d, rbf_field, ncolors=15, vmin=vmin, vmax=vmax)
mlab.points3d(puv[:, 0], puv[:, 1], 0 * puv[:, 0], scale_factor=0.1, color=(0, 0, 0))
f.scene.z_plus_view()
mlab.savefig(SAVE_DIR + "rbf_helmet_B.png", figure=f, magnification=4)

suh_field = (
    np.einsum("ijk,ij->ik", c.B_coupling(helmet.vertices), helmet.vertex_normals) @ s
)


f = plot_data_on_vertices(helmet2d, suh_field, ncolors=15, vmin=vmin, vmax=vmax)
mlab.points3d(puv[:, 0], puv[:, 1], 0 * puv[:, 0], scale_factor=0.1, color=(0, 0, 0))
f.scene.z_plus_view()
mlab.savefig(SAVE_DIR + "suh_helmet_B.png", figure=f, magnification=4)


Bca, Bcb = sphfield(helmet.vertices, lmax, normalization="energy", R=R)

# sph-components at sensors
sph_field = np.einsum("ijk,ij->ik", Bca, helmet.vertex_normals) @ alpha


f = plot_data_on_vertices(helmet2d, sph_field, ncolors=15, vmin=vmin, vmax=vmax)
mlab.points3d(puv[:, 0], puv[:, 1], 0 * puv[:, 0], scale_factor=0.1, color=(0, 0, 0))
f.scene.z_plus_view()
mlab.savefig(SAVE_DIR + "sph_helmet_B.png", figure=f, magnification=4)

#%% MNE interpolates using splines or something
#%% Compute potential
# U_sph = potential(
# p, alpha, np.zeros(alpha.shape), lmax=lmax, normalization="energy", R=R
# )
#
# U_suh = c.U_coupling(p) @ s

# evoked1 = evoked.copy()
# evoked1.data[:, :] = np.tile(U_sph.T, (evoked.times.shape[0], 1)).T
# evoked1.plot_topomap(times=0.080, ch_type="mag")

# evoked1 = evoked.copy()
# evoked1.data[:, :] = np.tile(U_suh.T, (evoked.times.shape[0], 1)).T
# evoked1.plot_topomap(times=0.080, ch_type="mag")


#%% interpolate data on planar mesh
from bfieldtools.utils import load_example_mesh
from bfieldtools.mesh_calculus import gradient

plane = load_example_mesh("10x10_plane_hires")
scaling_factor = 0.03
plane.apply_scale(scaling_factor)
# Rotate to x-plane
t = np.eye(4)
theta = np.pi / 2 * 1.2
t[1:3, 1:3] = np.array(
    [[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]]
)
plane.apply_transform(t)

c.U_coupling.reset()
U_suh = c.U_coupling(plane.vertices) @ a
# Adapt mesh to the function and calculate new points
for i in range(2):
    g = np.linalg.norm(gradient(U_suh, plane), axis=0)
    face_ind = np.flatnonzero(g > g.max() * 0.05)
    plane = plane.subdivide(face_ind)
    U_suh = c.U_coupling(plane.vertices) @ a

U_sph = potential(
    plane.vertices, alpha, np.zeros(alpha.shape), lmax=lmax, normalization="energy", R=R
)

#%%


# Mask inside/outside using solid angle
mask = abs(c.U_coupling.matrix.sum(axis=1)) < 1e-6
f = plot_data_on_vertices(plane, U_suh * mask, ncolors=15)
# plot_mesh(mesh, figure=f)
f = plot_data_on_vertices(plane, U_sph * mask, ncolors=15)
# plot_mesh(mesh, figure=f)
f = plot_data_on_vertices(plane, (U_suh - U_sph) * mask, ncolors=15)
plot_mesh(mesh, figure=f)



"""
Contains functions for computing thermal noise in conductive thin objects.

"""

__all__ = [
    "compute_current_modes",
    "noise_covar",
    "noise_covar_dir",
    "noise_var",
    "visualize_current_modes",
]

import numpy as np
import trimesh

from .suhtools import SuhBasis
from .mesh_conductor import MeshConductor
from .mesh_magnetics import magnetic_field_coupling


def compute_current_modes(
    obj,
    T,
    resistivity,
    thickness,
    mode="AC",
    freqs=np.array((0,)),
    Nmodes=None,
    return_eigenvals=False,
    **kwargs
):
    """
    Calculates the (AC or DC) Johnson noise current modes on the conducting surface.
    
    Parameters
    ----------
    obj: Trimesh-object or MeshConductor-object
        Represents the boundary on which current density is specified
        or MeshConductor object that wraps the mesh
    T: float
        Temperature in Kelvins
    resistivity: float or array (Nfaces)
        Resistivity value in Ohm/meter
    thickness: float or array (Nfaces)
        Thickness of the surface. NB! Must be small in comparison to observation distance
    mode: 'AC' or 'DC'
        Calculate modes as a function of frequency or just at DC?
    freqs: Nfreqs array
        The frequencies at which the eddy-current modes are computed. Obsolete for 
        'DC' mode. In 'AC', calculate at DC in default.
    Nmodes: int
        How many modes are computed? If None, all Nvertices modes are computed
    return_eigenvals: boolean
        Return also the eigenvalues (the inverse circuit time constants)?
    kwargs: dict
        Passed to Conductor creation if a Trimesh object is passed as 'obj'
        
    Returns
    -------
    vl: (Nvertices x Nmodes x Nfreqs) (AC) or (Nvertices x Nmodes) (DC) array
        The spectral eddy-current modes
    u: Nmodes array
        The eigenvalues
    """

    kB = 1.38064852e-23

    if isinstance(obj, MeshConductor):
        obj.resistivity = resistivity
        obj.thickness = thickness
        print(
            "Updated MeshConductor object thickness and resistivity attributes according to function parameters"
        )
    elif isinstance(obj, trimesh.Trimesh):
        obj = MeshConductor(
            mesh_obj=obj,
            resistivity=resistivity,
            thickness=thickness,
            resistance_full_rank=False,
            **kwargs
        )
    else:
        raise TypeError("obj type should be either Trimesh or Conductor")

    if mode == "AC":
        suh = SuhBasis(obj, Nc=Nmodes, magnetic="AC")
    elif mode == "DC":
        suh = SuhBasis(obj, Nc=Nmodes, magnetic="DC")
    else:
        raise ValueError("Mode should be either 'AC' or 'DC'")

    v = suh.basis
    u = suh.eigenvals

    Nfreqs = len(freqs)

    if mode == "AC":
        Nfreqs = len(freqs)
        vl = np.zeros((len(suh.mesh_conductor.mesh.vertices), v.shape[1], Nfreqs))

        # Scale the eigenmodes with the spectral density of the thermal noise current
        for i in range(v.shape[1]):
            amp = (
                2
                * np.sqrt(kB * T / u[i])
                * np.sqrt(1 / (1 + (2 * np.pi * freqs / u[i]) ** 2))
            )
            vl[:, i, :] = (
                (
                    np.zeros((Nfreqs, vl.shape[0]))
                    + suh.mesh_conductor.inner2vert @ v[:, i]
                ).T
            ) * amp
    elif mode == "DC":
        vl = np.zeros((len(suh.mesh_conductor.mesh.vertices), v.shape[1]))

        # Scale the eigenmodes with the spectral density of the thermal noise current
        for i in range(v.shape[1]):
            amp = 2 * np.sqrt(kB * T / u[i])
            vl[:, i] = suh.mesh_conductor.inner2vert @ v[:, i] * amp

    if return_eigenvals:
        return vl, u
    else:
        return vl


def noise_covar(B_coupling, vl, Nmodes=None):
    """
    Calculates (AC or DC) magnetic noise covariance along x, y and z from the 
    modes vl.
    
    Parameters
    ----------
    B_coupling: ndarray (Np, 3, Nvertices)
        Magnetic field coupling matrix from the mesh
    vl: ndarray (Nvertices, Nmodes, x Nfreqs) or (Nvertices, Nmodes)
        The Johnson noise current modes on the mesh
    Nmodes: int
        How many modes are included? If None, all modes in vl are included

    Returns
    -------
    Bcov: ndarray (Np, Np, 3, Nfreqs) or (Np, Np, 3)
        Magnetic noise covariance
    """
    if Nmodes is None:
        Nmodes = vl.shape[1]

    if vl.ndim == 2:
        b = np.einsum("ihj,jl->ilh", B_coupling, vl[:, 0:Nmodes])
        Bcov = np.einsum("jih,lih->jlh", b, b)
    else:
        b = np.einsum("ihj,jlk->ilhk", B_coupling, vl[:, 0:Nmodes, :])
        Bcov = np.einsum("jihk,lihk->jlhk", b, b)

    return Bcov


def noise_var(B_coupling, vl, Nmodes=None):
    """
    Calculates (AC or DC) magnetic noise variance along x, y and z from the 
    modes vl.
    
    Parameters
    ----------
    B_coupling: ndarray (Np, 3, Nvertices)
        Magnetic field coupling matrix from the mesh
    vl: ndarray (Nvertices, Nmodes, x Nfreqs) or (Nvertices, Nmodes)
        The Johnson noise current modes on the mesh
    Nmodes: int
        How many modes are included? If None, all modes in vl are included

    Returns
    -------
    Bcov: ndarray (Np, 3, Nfreqs) or (Np, 3)
        Magnetic noise variance
    """
    if Nmodes is None:
        Nmodes = vl.shape[1]

    if vl.ndim == 2:
        b = np.einsum("ihj,jl->ilh", B_coupling, vl[:, 0:Nmodes])
        Bcov = np.einsum("ijh,ijh->ih", b, b)
    else:
        b = np.einsum("ihj,jlk->ilhk", B_coupling, vl[:, 0:Nmodes, :])
        Bcov = np.einsum("ijhk,ijhk->ihk", b, b)

    return Bcov


def noise_covar_dir(B_coupling, vl, Nmodes=None):
    """
    Calculates (AC or DC) magnetic noise covariance between x, y and z directions
    from the modes vl.
    
    Parameters
    ----------
    B_coupling: ndarray (Np, 3, Nvertices)
        Magnetic field coupling matrix from the mesh
    vl: ndarray (Nvertices, Nmodes, x Nfreqs) or (Nvertices, Nmodes)
        The Johnson noise current modes on the mesh
    Nmodes: int
        How many modes are included? If None, all modes in vl are included

    Returns
    -------
    Bcov: ndarray (Np, 3, 3, Nfreqs) or (Np, 3, 3)
        Magnetic noise covariance x, y and z field components
    """
    if Nmodes is None:
        Nmodes = vl.shape[1]

    if vl.ndim == 2:
        b = np.einsum("ihj,jl->ilh", B_coupling, vl[:, 0:Nmodes])
        Bcov = np.einsum("ihj,ihl-> ijl", b, b)
    else:
        b = np.einsum("ihj,jlk->ilhk", B_coupling, vl[:, 0:Nmodes, :])
        Bcov = np.einsum("ihjk,ihlk-> ijlk", b, b)

    return Bcov


def sensornoise_covar(mesh, p, n, w, vl, Nmodes=None):
    """
    Calculates the upper diagonal of (AC or DC) magnetic noise covariance
    on a sensor array described by integration points.
    
    Assumes same number of integration points for each sensor.
    
    Parameters
    ----------
    mesh: Trimesh-object
        The boundary on which current density is specified
    p: ndarray (Nsensors, Np, 3)
        Coordinates of the integration points of the Nsensors
    n: ndarray (Nsensors, Np, 3)
        Orientations of the integration points of the Nsensors
    w: ndarray (Nsensors, Np)
        The weights of the integration points
    vl: ndarray (Nvertices, Nmodes, x Nfreqs) or (Nvertices, Nmodes)
        The Johnson noise current modes on the mesh
    Nmodes: int
        How many modes are included? If None, all modes in vl are included

    Returns
    -------
    Bcov: ndarray (Nsensors, Nsensors, Nfreqs) or (Nsensors, Nsensors)
        Magnetic noise covariance
    """
    if Nmodes is None:
        Nmodes = vl.shape[1]

    Nsensors = p.shape[0]
    Np = p.shape[1]

    # Compute the magnetic field coupling matrices along the orientations
    # of the integration points
    b = np.zeros((Np, Nsensors, mesh.vertices.shape[0]))
    for i in range(Nsensors):
        B_coupling = magnetic_field_coupling(mesh, p[i], analytic=True)
        b[:, i] = np.einsum("ijk,ij->ik", B_coupling, n[i])

    # Compute the magnetic noise covariance on the sensor array using the
    # weights of the integration points
    if vl.ndim == 2:
        Bcov = np.zeros((Nsensors, Nsensors))
        for i in range(Nsensors):
            for j in range(i, Nsensors):
                Bcov[i, j] = (
                    w[i].T
                    @ b[:, i]
                    @ vl[:, 0:Nmodes]
                    @ vl[:, 0:Nmodes].T
                    @ b[:, j].T
                    @ w[j]
                )
    else:
        Bcov = np.zeros((Nsensors, Nsensors, vl.shape[2]))
        for i in range(Nsensors):
            for j in range(i, Nsensors):
                temp = w[i].T @ b[:, i]
                bi = np.einsum("ij,jkh->ikh", temp[None, :], vl[:, 0:Nmodes, :])
                temp = w[j].T @ b[:, j]
                bj = np.einsum("ij,jkh->ikh", temp[None, :], vl[:, 0:Nmodes, :])

                Bcov[i, j] = np.einsum("ijh,ijh->h", bi, bj)
    return Bcov


def visualize_current_modes(
    mesh, vl, Nmodes, scale, contours=True, colormap="bwr", dist=0.5
):
    """
    Visualizes current modes up to Nmodes.

    Parameters
    ----------
    mesh: Trimesh mesh object
        The surface mesh
    vl: Nvertices x Nvertices array
        The normalized eddy-current modes vl[:,i]
    Nmodes: int
        Number of modes to be plotted
    scale: float
        Scaling factor
    contours: boolean
        If True, show contours
    colormap: string
        Which (matplotlib) colormap to use

    """
    from mayavi import mlab

    N1 = np.floor(np.sqrt(Nmodes))
    dx = (mesh.vertices[:, 0].max() - mesh.vertices[:, 0].min()) * (1 + dist)
    dy = (mesh.vertices[:, 1].max() - mesh.vertices[:, 1].min()) * (1 + dist)

    i = 0
    j = 0
    for n in range(Nmodes):
        print(i, j)
        points = mesh.vertices.copy()
        points[:, 0] += i * dx
        points[:, 1] += j * dy
        s = mlab.triangular_mesh(
            *points.T, mesh.faces, scalars=vl[:, n], colormap=colormap
        )

        limit = np.max(np.abs(vl[:, n]))

        s.module_manager.scalar_lut_manager.number_of_colors = 256
        s.module_manager.scalar_lut_manager.data_range = np.array([-limit, limit])
        s.actor.mapper.interpolate_scalars_before_mapping = True
        s.enable_contours = contours

        if i < N1:
            i += 1
        else:
            j += 1
            i = 0

    return s



"""
Contains a class used for wrapping a mesh (in the form of a Trimesh object) together with
some convenient functions and properties.

"""

__all__ = [
    "CouplingMatrix",
    "MeshConductor",
    "StreamFunction",
    "matrixwrapper",
    "save_pickle",
    "load_pickle",
]

from time import time
import pickle

import trimesh
import numpy as np
from scipy.sparse import issparse

from . import utils
from .mesh_calculus import laplacian_matrix, mass_matrix
from .mesh_impedance import (
    self_inductance_matrix,
    resistance_matrix,
    mutual_inductance_matrix,
)
from .mesh_magnetics import (
    magnetic_field_coupling,
    scalar_potential_coupling,
    vector_potential_coupling,
)
from .suhtools import SuhBasis
from .sphtools import compute_sphcoeffs_mesh
from .viz import plot_mesh, plot_data_on_vertices
from .contour import scalar_contour
from .line_conductor import LineConductor


def matrixwrapper(func):
    """
    Wrapper for lazy computation of MeshConductor matrices with basis change
    """

    def wrapper(obj):
        name = func.__name__
        M = obj.matrices[name[1:]]
        if M is None:
            print("Computing the %s matrix..." % str(name[1:]))
            M = obj.matrices[name[1:]] = func(obj)
        return obj.basis.T @ M @ obj.basis

    return wrapper


class MeshConductor:
    """
    Class that is used for surface mesh field calculations.
    
    Computation functions are typically external functions that are called
    using lazy properties.

    The mesh surface can consist of a single contiguous surface or several separate
    surfaces within a single mesh object. The MeshConductor object can handle data defined
    on the mesh being represented in several different bases:
     - inner (default)
     - vertex:
     - suh: surface harmonics basis. Order given by N_suh

    The bases can include built-in boundary conditions for the data: inner and
    suh bases assume dirichlet boundary condition (equal value within each boundary),
    while vertex basis does not set a boundary condition.
    
    """

    def __init__(
        self,
        verts=None,
        tris=None,
        mesh_file=None,
        mesh_obj=None,
        process=False,
        fix_normals=True,
        basis_name="inner",
        resistivity=1.68 * 1e-8,
        thickness=1e-4,
        **kwargs
    ):
        """
        First priority is to use given Trimesh object (mesh_obj).
        Second priority is to load mesh from file (mesh_file).
        Third priority is to use given verts and tris arrays (verts, tris).

        Parameters
        ----------
        verts: array-like (Nv, 3)
            Array of mesh vertices
        tris: array-like  (Nt, 3)
            Array of mesh faces
        mesh_obj: Trimesh mesh object
        mesh_file: string
            String describing the file path of a mesh file.
        process: boolean
            If True,  Trimesh will pre-process the mesh.
        fix_normals: boolean
            If True,  normals+winding should be set so that they always point "out" from the origin.
        basis_name: string
            Which basis to use, must be 'inner', 'vertex' or 'suh'. See class docstring
        Resistivity: float or array (Nfaces)
            Resistivity value in Ohm/meter
        Thickness: float or array (Nfaces)
            Thickness of surface. NB! Must be small in comparison to observation distance
        kwargs:
            Additional options with default settings are:
                'outer_boundaries':None,
                'resistance_full_rank': False,
                'inductance_nchunks':None,
                'basis_name':'inner' (other: suh, vertex)
                'N_suh': 100
                'sph_normalization': 'default'
                'sph_radius': 1
                'N_sph': 5
                'approx_far': True
                'approx_far_margin': 2
             
            
        Notes
        ------
        
        **outer_boundaries**
        int or array_like, indices of outer boundaries given by utils.find_boundaries(). 
        One boundary index per mesh component. If None, outer_boundaries are set to 
        the longest boundary in each mesh component. When using basis 'inner', the outer boundary
        vertex values are fixed to zero.
        
        **resistance_full_rank** (Boolean)
        If True, applies inflation to the resistance matrix in order to increase
        the rank by one. By default, is False. Don't set to True without knowing what you are doing.
        
        **inductance_nchunks** (int or None)
        Number of serial chunks to split self-inductance computation into, saving memory but taking more time.
        When approx_far is True, using more chunks is more efficient (multiply by 10-100x)
        If None (default), attempts to set number of chunks automatically based on the amount of free memory.
        Unfortunately, this estimation is not perfect.

        **N_suh**
        Number of surface harmonics to use if basis_name is 'suh'
        
        **sph_normalization**
        'default' (Ylm**2 integrated over solid angle to 1) or
        'energy' (field energy of basis fields normalized to 1 in R-ball)
        
        **sph_radius**
        If sph_normalization is 'energy', defines the radius of the inner expansion
        
        **N_sph**
        Number of spherical harmonics degrees (l-degrees) to use for the spherical harmonics coupling computation
        
        **approx_far** (Boolean)
        If True, usesimple quadrature for points far from the source triangles when computing self-inductance
        
        **approx_far_margin** (non-negative float)
        Cut-off distance for "far" points measured in mean triangle side length.


        """
        #######################################################################
        # Load mesh, do mesh pre-processing
        #######################################################################

        if (
            mesh_obj
        ):  # First, if Trimesh object is given as parameter then use that directly
            self.mesh = mesh_obj

        elif (
            mesh_file
        ):  # Second, check if mesh_file passed and load mesh as Trimesh object
            self.mesh = trimesh.load(mesh_file, process=process)

        else:  # Fall back on verts, tris parameters
            if isinstance(verts, type(None)) or isinstance(tris, type(None)):
                raise ValueError(
                    "You must provide either verts and tris, a mesh object or a mesh file"
                )

            self.mesh = trimesh.Trimesh(verts, tris, process=process)

        if fix_normals:
            self.mesh = utils.fix_normals(self.mesh)

        #######################################################################
        # Apply options
        #######################################################################

        # Populate options dictionary with defaults if not specified
        self.opts = {
            "outer_boundaries": None,
            "resistance_full_rank": False,
            "inductance_nchunks": None,
            "N_suh": 100,
            "sph_normalization": "default",
            "sph_radius": 1,
            "N_sph": 5,
            "inductance_quad_degree": 2,
            "approx_far": True,
            "approx_far_margin": 2,
        }

        for key, val in kwargs.items():
            self.opts[key] = val

        #######################################################################
        # Set up holes/boundaries
        #######################################################################

        self.boundaries, self.inner_vertices = utils.find_mesh_boundaries(self.mesh)
        self.set_holes(self.opts["outer_boundaries"])

        #######################################################################
        # Set up physical properties and coupling matrices
        #######################################################################

        # Matrices in inner-weight basis
        self.matrices = {
            "laplacian": None,
            "mass": None,
            "inductance": None,
            "resistance": None,
        }

        self.resistivity = resistivity
        self.thickness = thickness

        self.B_coupling = CouplingMatrix(self, magnetic_field_coupling)
        self.U_coupling = CouplingMatrix(self, scalar_potential_coupling)
        self.A_coupling = CouplingMatrix(self, vector_potential_coupling)

        # Coupling to spherical harmonic field coefficients
        self._alpha_coupling = None
        self._beta_coupling = None

        #######################################################################
        # Set up stream function basis
        #######################################################################

        self.inner2vert = utils.inner2vert(self.mesh, self.inner_vertices, self.holes)
        self.vert2inner = utils.vert2inner(self.mesh, self.inner_vertices, self.holes)

        # Sets basis for first time, calling self.set_basis()
        self.basis_name = basis_name
        self.set_basis(self.basis_name)

    def set_basis(self, basis_name):
        """ The data is stored in vertex basis i.e. every
            element corresponds to one vertex. The basis matrix changes basis
            of the operators so that a coefficient vector in the desired
            basis can be multiplied directly with the operator

            basis_names : str 'vertex', 'inner' or 'suh'
        """
        from scipy.sparse import spdiags

        self.basis_name = basis_name

        if self.basis_name == "suh":
            # SUH basis with dirichlet boundary conditions
            self.suh_basis = SuhBasis(
                self.mesh, self.opts["N_suh"], boundary_condition="dirichlet"
            )
            self.basis = self.inner2vert @ self.suh_basis.basis
        elif self.basis_name == "inner":
            self.basis = self.inner2vert
        elif self.basis_name == "vertex":
            N = len(self.mesh.vertices)
            self.basis = spdiags(np.ones(N), 0, N, N)
        else:
            raise ValueError("streamfunction_basis must inner, vertex or suh")

    def set_holes(self, outer_boundaries=None):
        """ Set indices of holes to self.holes

            outer_boundaries: int or array_like, indices of outer boundaries in
                    self.boundaries. One boundary index per mesh component.
                    If None, outer_boundaries are set the longest
                    boundary in each mesh component
        """
        if len(self.boundaries) == 0:
            # The mesh is watertight
            self.holes = []
            return
        if outer_boundaries is None:
            # Have to determine if there multiple bodies and label
            # the boundaries according to them
            comps = trimesh.graph.connected_components(self.mesh.edges)
            b_labels = np.zeros(len(self.boundaries))
            for m, b in enumerate(self.boundaries):
                for n, c in enumerate(comps):
                    if b[0] in c:
                        b_labels[m] = n
            b_lengths = np.array([len(b) for b in self.boundaries])
            # Determine outer_boundaries by finding the boundary of max length
            # for each component
            outer_boundaries = []
            b_inds = np.arange(len(self.boundaries))
            for n in range(len(comps)):
                mask = b_labels == n
                outer_boundaries.append(
                    b_inds[mask][np.argwhere(b_lengths[mask] == b_lengths[mask].max())]
                )

        self.opts["outer_boundaries"] = outer_boundaries
        hole_inds = list(
            np.setdiff1d(np.arange(len(self.boundaries)), outer_boundaries)
        )
        if len(hole_inds) == 0:
            # The non-watertight meshes contain no holes
            self.holes = []
        else:
            self.holes = [self.boundaries[i] for i in hole_inds]

    @property
    def laplacian(self):
        """
        Surface laplacian matrix, returned in appropiate basis.

        For further information, see mesh_calculus.laplacian_matrix

        property-decorated wrapper.
        """
        return self._laplacian()

    @matrixwrapper
    def _laplacian(self):
        """
        Compute and return surface laplacian matrix.

        """
        laplacian = laplacian_matrix(self.mesh, None)
        return laplacian

    @property
    def mass(self):
        """
        Mass matrix, returned in appropiate basis.

        For further information, see mesh_calculus.mass_matrix

        property-decorated wrapper.

        """
        return self._mass()

    @matrixwrapper
    def _mass(self):
        """
        Compute and return mesh mass matrix.

        """
        mass = mass_matrix(self.mesh, lumped=False)
        return mass

    @property
    def inductance(self):
        """
        Self-inductance matrix, returned in appropiate basis.

        For further information, see mesh_impedance.self_inductance_matrix

        property-decorated wrapper.
        """
        return self._inductance()

    @matrixwrapper
    def _inductance(self):
        """
        Compute and return self-inductance matrix.

        """

        start = time()

        inductance = self_inductance_matrix(
            self.mesh,
            Nchunks=self.opts["inductance_nchunks"],
            quad_degree=self.opts["inductance_quad_degree"],
            approx_far=self.opts["approx_far"],
            margin=self.opts["approx_far_margin"],
        )

        duration = time() - start
        print("Inductance matrix computation took %.2f seconds." % duration)

        return inductance

    @property
    def resistance(self):
        """
        Resistance matrix. For further information, see mesh_impedance.resistance_matrix

        property-decorated wrapper.
        """
        return self._resistance()

    @matrixwrapper
    def _resistance(self):
        """
        Back-end of resistance matrix computation
        """
        sheet_resistance = self.resistivity / self.thickness
        resistance = resistance_matrix(self.mesh, sheet_resistance).toarray()

        # Compensate for rank n-1 by adding offset, otherwise this
        # operator map constant vectors to zero
        if self.opts["resistance_full_rank"]:
            scale = np.mean(sheet_resistance)
            resistance += np.ones(resistance.shape) / resistance.shape[0] * scale

        return resistance

    def mutual_inductance(self, mesh_conductor_other, quad_degree=1, approx_far=True):
        """
        Mutual inductance between this MeshConductor object and another

        Parameters:
            mesh_conductor_other: MeshConductor object

        Returns:
            M: mutual inductance matrix M(self, other) in
                in the bases specified in the mesh_conductor object

        """
        M = mutual_inductance_matrix(
            self.mesh,
            mesh_conductor_other.mesh,
            quad_degree=quad_degree,
            approx_far=approx_far,
        )
        # Convert to the desired basis
        M = self.basis.T @ M @ mesh_conductor_other.basis

        return M

    def set_sph_options(self, **kwargs):
        """
        Set or reset options related to the spherical harmonics (SPH) coupling of the 
        MeshConductor object. If SPH couplings have already been computed, these will be 
        flushed.
    

        Parameters
        ----------
        **kwargs 
            Any parameter related to sph, namely "N_sph", "sph_normalization", "sph_radius" 

        Returns
        -------
        None.

        """

        # Flush old results
        self._alpha_coupling = None
        self._beta_coupling = None

        # Assign new values to opts, filter out opts not related to sph
        for key, val in kwargs.items():
            if key in ("N_sph", "sph_normalization", "sph_radius"):
                self.opts[key] = val

    @property
    def sph_couplings(self):
        """
        Spherical harmonic mappings from a stream function defined on the
        MeshConductor mesh.

        property-decorated wrapper
        """
        return self._sph_couplings()

    def _sph_couplings(self):
        """
        Compute spherical harmonic mappings and store them for further use
        """
        if self._alpha_coupling is None:
            print("Computing coupling matrices")
            Calpha, Cbeta = compute_sphcoeffs_mesh(
                self.mesh,
                self.opts["N_sph"],
                self.opts["sph_normalization"],
                self.opts["sph_radius"],
            )
            # Store the results for further use
            self._alpha_coupling = Calpha
            self._beta_coupling = Cbeta
        else:
            Calpha = self._alpha_coupling
            Cbeta = self._beta_coupling

        return Calpha @ self.basis, Cbeta @ self.basis

    def __setattr__(self, name, value):
        """
        Modified set-function to take into account post-hoc changes to e.g. resistance
        """
        self.__dict__[name] = value

        # If resistance-affecting parameter is changed after the resistance matrix has been computed,
        # then flush old result and re-compute
        if (name in ("resistivity", "thickness")) and self.matrices[
            "resistance"
        ] is not None:
            self.matrices["resistance"] = None  # Flush old matrix
            self._resistance()  # Re-compute with new parameters

        if (name in ("resistivity", "thickness")) and self.matrices[
            "resistance"
        ] is not None:
            self.matrices["resistance"] = None  # Flush old matrix
            self._resistance()  # Re-compute with new parameters

    def plot_mesh(self, cull_front=False, cull_back=False, **kwargs):
        """
        Simply plot the mesh surface in mayavi. kwargs are passed to
        viz.plot_mesh

        """

        return plot_mesh(
            self.mesh, cull_front=cull_front, cull_back=cull_back, **kwargs
        )

    def save_pickle(self, target_file):
        """
        Save the MeshConductor object using a pickled Python file

        Parameters
        ----------
        target_file: str
            File name or file object to save to

        """

        pickle.dump(obj=self, file=open(target_file, "wb"))


def save_pickle(obj, target_file):
    """
    Save the MeshConductor object using a pickled Python file

    Parameters
    ----------
    obj: object to save to file
    target_file: str
        file name or file object to save to

    """

    pickle.dump(obj=obj, file=open(target_file, "wb"), protocol=-1)


def load_pickle(target_file):
    """
    Load pickled MeshConductor object from file

    Parameters
    -----------
    target_file: str
        File name or file object to load from

    Returns
    -------
    obj: loaded MeshConductor object


    """

    obj = pickle.load(open(target_file, "rb"))

    return obj


class CouplingMatrix:
    """
    General-use class that contains a data array (a coupling matrix)
    and a bookkeeping list of computed points.

    When called, returns the coupling matrix for queried points.
    If some output has already been computed, use pre-computed values instead
    and only compute missing parts.


    """

    def __init__(self, parent, function):

        # Bookkeeping array, which points are already computed
        # Indexed in same order as the matrix
        self.points = np.array([])

        self.matrix = np.array([])

        self.parent = parent
        self.function = function

    def reset(self):
        """ Reset the coupling matrix and points
        """
        self.points = np.array([])
        self.matrix = np.array([])

    def __call__(self, points, s=None, *fun_args, **kwargs):
        """
        Returns the output of self.function(self.parent, points).
        If some output has already been computed, use pre-computed values instead
        and only compute missing parts.

        Parameters
        ----------
            points: (N_points, ... ) numpy array
                Array containing query points
            s: (Nbasis,) numpy array
                If None, return coupling matrix, 
                if not None, return field, i.e, coupling_matrix @ s
                default None
            *fun_args: additional arguments
                Optional, additional arguments that are passed to self.function

        Returns
        -------
            (N_points, ...) numpy array

        """

        if len(self.points) == 0:
            # Nothing calculated,yet
            self.matrix = self.function(self.parent.mesh, points, *fun_args, **kwargs)
            self.points = points

            M = self.matrix

        else:
            # Check which points exist and which need to be computed
            p_existing_point_idx, m_existing_point_idx = np.where(
                (self.points == points[:, None]).all(axis=-1)
            )
            missing_point_idx = np.setdiff1d(
                np.arange(0, len(points)), p_existing_point_idx
            )

            # If there are missing points, compute and add
            if len(missing_point_idx) > 0:
                missing_points = points[missing_point_idx]

                new_matrix_elems = self.function(
                    self.parent.mesh, missing_points, *fun_args, **kwargs
                )

                # Append newly computed point to coupling matrix, update bookkeeping
                self.points = np.vstack((self.points, missing_points))
                self.matrix = np.vstack((self.matrix, new_matrix_elems))

                # Re-compute indices of queried points, now that all should exist
                p_existing_point_idx, m_existing_point_idx = np.where(
                    (self.points == points[:, None]).all(axis=-1)
                )

            M = self.matrix[m_existing_point_idx]

        if s is None:
            if self.matrix.ndim == 2:
                M = M @ self.parent.basis

            elif self.matrix.ndim == 3:

                # Handle both sparse and dense basis matrices quickly
                if issparse(self.parent.basis):
                    Mnew = []
                    for n in range(3):
                        Mnew.append(M[:, n, :] @ self.parent.basis)
                    M = np.swapaxes(np.array(Mnew), 0, 1)
                else:
                    M = np.einsum("ijk,kl->ijl", M, self.parent.basis)

            else:
                raise ValueError("Matrix dimensions not ok")

            return M
        else:
            if self.matrix.ndim == 2:
                field = M @ (self.parent.basis @ s)
            elif self.matrix.ndim == 3:
                ss = self.parent.basis @ s
                field = np.einsum("ijk,k->ij", M, ss)
            else:
                raise ValueError("Matrix dimensions not ok")

            return field


class StreamFunction(np.ndarray):
    """ Class for representing stream function(s) on a MeshConductor

        Handles the mapping between different bases, e.g. inner vertices <->
        all vertices <-> surface harmonics

        Parameters:
            vals: array of shape (N,) or (N,M)
                where N corresponds to
                the number of inner vertices in the mesh_conductor or the
                the number of all vertices in the MeshConductor.

                Multiple (M) stream functions can be stored in the object
                by specifying vals with shape (N,M)
            mesh_conductor:
                MeshConductor object
    """

    def __new__(cls, input_array, mesh_conductor=None):
        #        print('In __new__ with class %s' % cls)
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        obj = np.asarray(input_array).view(cls)
        # add the new attribute to the created instance
        obj.mesh_conductor = mesh_conductor

        return obj

    def __init__(self, input_array, mesh_conductor=None):
        #        print('In __init__ with class %s' % self.__class__)

        self.basis_name = self.mesh_conductor.basis_name
        self.basis = self.mesh_conductor.basis

        self.inner2vert = self.mesh_conductor.inner2vert
        self.vert2inner = self.mesh_conductor.vert2inner

    def __array_finalize__(self, obj):
        #
        #        print('In array_finalize:')
        #        print('   self type is %s' % type(self))
        #        print('   obj type is %s' % type(obj))

        if obj is None:
            #            obj = obj.view(np.ndarray)
            return

        self.mesh_conductor = getattr(obj, "mesh_conductor", None)

    def __array_wrap__(self, out_arr, context=None):
        """
        Return a ndarray for all ufuncs, since the StreamFunction attributes no longer apply
        when e.g. changing shape etc
        """
        return np.ndarray.__array_wrap__(self, out_arr, context).view(np.ndarray)

    def __getitem__(self, k):
        return np.ndarray.__getitem__(self, k).view(np.ndarray)

    def __repr__(self):
        return "%s, %s,  basis: %s" % (
            self.__class__.__name__,
            self.view(np.ndarray).__repr__(),
            self.basis_name,
        )

    @property
    def vert(self):
        """
        Returns the stream function in vertex basis
        """
        return self.basis @ self

    @property
    def inner(self):
        """
        Returns the stream function in inner basis
        """
        if self.basis_name == "inner":
            return self

        return self.vert2inner @ self.basis @ self

    @property
    def power(self):
        """
        Stream-function resistive power
        """

        # Compute resistance matrix if not present
        if self.mesh_conductor.matrices["resistance"] is None:
            self.mesh_conductor.resistance

        R = self.mesh_conductor.matrices["resistance"]  # Always in vertex basis

        return self.T @ self.basis.T @ R @ self.basis @ self

    @property
    def magnetic_energy(self):
        """
        Stream-function magnetic energy
        """

        # Compute inductance matrix if not present
        if self.mesh_conductor.matrices["inductance"] is None:
            self.mesh_conductor.inductance

        M = self.mesh_conductor.matrices["inductance"]  # Always in vertex basis

        return 0.5 * self.T @ self.basis.T @ M @ self.basis @ self

    def coil_inductance(self, Nloops):
        """
        

        Parameters
        ----------
        Nloops : the number of wire loops in the discrete coil (int)

        Returns
        -------
        inductance: scalar (float)

        """
        scaling = Nloops / (self.vert.max() - self.vert.min())
        L_approx = 2 * self.magnetic_energy * (scaling ** 2)

        return L_approx

    def plot(self, background=True, contours=False, **kwargs):
        """
        Plot the stream function
        """
        from mayavi import mlab

        mesh = self.mesh_conductor.mesh
        scalars = self.vert
        if "vmin" not in kwargs.keys():
            kwargs["vmin"] = -np.max(abs(scalars))
        if "vmax" not in kwargs.keys():
            kwargs["vmax"] = np.max(abs(scalars))

        s = plot_data_on_vertices(mesh, scalars, **kwargs)
        if contours:
            s.enable_contours = True
            s.contour.number_of_contours = contours
            if background:
                mlab.triangular_mesh(
                    *mesh.vertices.T, mesh.faces, color=(0.5, 0.5, 0.5), opacity=0.2
                )

        return s

    def discretize(self, N_contours=10, contours=None):
        """
        Wrapper method for contour.scalar_contour, turns the piecewise linear
        stream function into isolines/contours in the form of polylines.

        Parameters
        ----------
        N_contours: int
            Number of contours to generate
        contours: array-like
            Optional argument for manual input of contour levels. Overrides `N_contours`

        Returns
        -------
        contour_polys: list
            list with length `N_contours`. Each list element is anumpy array containing the
            coordinats of each polygon vertex.
        contour_values: array-like
            Vector containing the scalar function value for each contour line
        """
        return LineConductor(
            mesh=self.mesh_conductor.mesh,
            scalars=self.vert,
            N_contours=N_contours,
            contours=contours,
        )



"""
Functions for working with current polylines, e.g. for calculating the magnetic field and potentials as well as the inductance

"""

__all__ = [
    "magnetic_field",
    "mutual_inductance",
    "scalar_potential",
    "self_inductance",
    "vector_potential",
]

import numpy as np
from .integrals import omega
import trimesh


def cross(r1, r2):
    """ Cross product without overhead
    """
    result = np.zeros(r1.shape)
    result[0] = r1[1] * r2[2] - r1[2] * r2[1]
    result[1] = r1[2] * r2[0] - r1[0] * r2[2]
    result[2] = r1[0] * r2[1] - r1[1] * r2[0]
    return result


def magnetic_field(vertices, points):
    """ Compute B-field of a segmented line current.
        See: Compact expressions for the Biot–Savart fields of a filamentary segments
        by Hanson & Hirshman: https://doi.org/10.1063/1.1507589


        Parameters
        ----------

        vertices: (N_line, 3) array
            Vertices of the line with N_line-1 segments
            The first and last vertices should be the same to close the loop.
        points:   (N_points, 3) array
            Magnetic field evaluation points

        Returns
        -------
        bfield: (N_points, 3) array
            Magnetic field at evaluation points

    """
    field = np.zeros(points.T.shape)
    for i in range(len(vertices) - 1):
        r1 = vertices[i]
        r2 = vertices[i + 1]

        # Vectors between vertices and field points
        a1 = points.T - r1.reshape(3, 1)
        a2 = points.T - r2.reshape(3, 1)

        # Direction of the field
        f = cross(a1, a2)

        # Vector lengths
        d1 = np.sqrt(np.sum(a1 ** 2, axis=0))
        d2 = np.sqrt(np.sum(a2 ** 2, axis=0))

        # Normalize direction field and divide by cylindrical distance
        f *= (d1 + d2) / (d1 * d2 * (d1 * d2 + np.sum(a1 * a2, axis=0)))

        field = field + f

    return field.T * 1e-7


def vector_potential(vertices, points, reg=1e-12, symmetrize=True):
    """ Compute vector potential of a segmented line currents.
        Based on straightforward integration of 1/r potential over a line
        i.e. the gamma0 integral

        See: Compact expressions for the Biot–Savart fields of a filamentary segments
        by Hanson & Hirshman: https://doi.org/10.1063/1.1507589


        Parameters
        ----------
        vertices: (N_line, 3) array
            Vertices of the line with N_line-1 segments
            The first and last vertices should be the same to close the loop.
        points: (N_points, 3) array
            Evaluation points

        Returns
        -------
        A: array (Npoints, 3)
            Vector potential

    """

    segments = vertices[1:] - vertices[:-1]
    RR = vertices[:, None, :] - points[None, :, :]
    dotprods2 = np.sum(RR[1:] * segments[..., None, :], axis=-1)
    dotprods1 = np.sum(RR[:-1] * segments[..., None, :], axis=-1)
    ss = np.linalg.norm(segments, axis=-1)
    segments /= ss[..., None]
    rr = np.linalg.norm(RR, axis=-1)

    # Regularize s.t. neither the denominator or the numerator can be zero
    # Avoid numerical issues directly at the edge
    res = np.log(
        (rr[1:] * ss[..., None] + dotprods2 + reg)
        / (rr[:-1] * ss[..., None] + dotprods1 + reg)
    )

    # Symmetrize the result since on the negative extension of the edge
    # there's division of two small values resulting numerical instabilities
    # (also incompatible with adding the reg value)
    if symmetrize:
        res2 = -np.log(
            (rr[1:] * ss[..., None] - dotprods2 + reg)
            / (rr[:-1] * ss[..., None] - dotprods1 + reg)
        )
        res = np.where(dotprods1 + dotprods2 > 0, res, res2)

    return 1e-7 * np.sum(res[..., None] * segments[..., None, :], axis=0)


def scalar_potential(vertices, points):
    """ Computes the scalar magnetic potential of a segmented current loop at given points.
        This is equal to the solid angle spanned by the loop (polygon), times a constant.
        The first and last vertices should be the same to close the loop.


        Parameters
        ----------
        vertices: (N_line, 3) array
            Vertices of the line with N_line-1 segments
            The first and last vertices should be the same to close the loop.
        points: (N_points, 3) array
            Evaluation points

        Returns
        -------
        U: array (Npoints, )
            Scalar magnetic potential

    """

    N_verts = len(vertices)

    # VERTEX MASS CENTRE
    mass_center = np.mean(vertices, axis=0)

    vertices = np.vstack((vertices, mass_center))

    # CREATE TRIANGLE FAN
    faces = np.full(shape=(N_verts - 1, 3), fill_value=np.nan, dtype=int)

    for i in range(N_verts - 1):
        faces[i] = np.array([i, i + 1, N_verts])

    R1 = vertices[faces]
    R2 = points

    RR = R2[:, None, None, :] - R1[None, :, :, :]

    # COMPUTE SOLID ANGLE, DIVIDE BY 4*PI*mu0
    return np.sum(omega(RR), axis=1) * 1e-7


def mutual_inductance(path1, path2, Nquad=2, radius=1e-3):
    """ Compute magnetic flux created by a segmented line current loops
        (path1) on a another closed loop of segmented current
        (path2). The other loop(s) is numerically integrated.

        In other words, calculate mutual inductance of the current loops.

        **SELF INDUCTANCE OF A LOOP IS BASED ON ROUND-WIRE APPROXIMATION**
        
        Parameters
        ----------
        path1: trimesh.Path3D-object
        
        path2: trimesh.Path3D-object

        Nquad: int, the number of quadrature points on line segments
        
        radius: float
            radius of the wire, only used for self-inductance calcultion
            self-inductance is dependent on the wire cross section
        Returns
        -------
         flux in the other loop generated by loops (Nloops,)


    """
    # Calculate quadrature points linearly spaced on the line segments
    # of the other loop
    t = np.linspace(0, 1, Nquad + 1)

    fluxes = np.zeros((len(path1.entities), len(path2.entities)))
    for j, loop2 in enumerate(path2.entities):
        vertices_other = path2.vertices[loop2.nodes, :]
        # sides = vertices_other[1:] - vertices_other[:-1]
        sides = vertices_other[:, 1] - vertices_other[:, 0]
        segments = (t[1:, None, None] - t[:-1, None, None]) * sides

        tq = 0.5 * (t[1:] + t[:-1])  # Nquad points on the segments

        # Nquad, Nsegments, 3
        points = vertices_other[:, 0] + tq[:, None, None] * sides
        shape = points.shape

        for i, loop1 in enumerate(path1.entities):
            if (path1 is path2) and i == j:
                fluxes[i, j] = self_inductance(path1, loop1, radius)
            else:
                vertices = path1.vertices[loop1.points]
                a = vector_potential(vertices, points.reshape(-1, 3))

                # Nquad, Nsegments, 3
                a = a.reshape(shape)

                # Take dot product between vector potential and the line segments
                # corresponding to each quadrature points (axis=0) and sum over the
                # segments (axis=1) and quadrature points on each segment (axis=2)

                fluxes[i, j] = np.sum(a * segments, axis=(0, 1, 2))

    return fluxes


def self_inductance(path, loop, radius=1e-3):
    """
    Calculate self inductance based on round-wire approximation
    http://www.thompsonrd.com/induct2.pdf
    section 5.
    

    Parameters
    ----------
    path : path3D object
    loop : entity of the path3D object
        entity describing the single loop
    radius : the radius of the wire
        DESCRIPTION. The default is 1e-3.

    Returns
    -------
    self_ind : float

    """
    mu0_over2pi = (1e-7) * 2
    # Perimeter
    p = loop.length(path.vertices)

    # CREATE TRIANGLE FAN for area calculation
    vertices = path.vertices[loop.points]
    N_verts = len(vertices)
    mass_center = np.mean(vertices, axis=0)
    vertices = np.vstack((vertices, mass_center))
    faces = np.full(shape=(N_verts - 1, 3), fill_value=np.nan, dtype=int)
    for i in range(N_verts - 1):
        faces[i] = np.array([i, i + 1, N_verts])
    mesh = trimesh.Trimesh(vertices, faces)
    # Area
    A = mesh.area

    # This is an approximation for round wire
    self_ind = mu0_over2pi * p * (np.log(2 * A / (p * radius)) + 0.25)

    return self_ind



"""
Spherical harmonics-generating coil design
==========================================

Example showing a basic biplanar coil producing a field profile defined by
spherical harmonics. We use the surface harmonics basis for the stream function,
and optimize the coupling to spherical harmonics components, thus creating a compact
optimization problem that can be solved very quickly.

"""

import numpy as np
import trimesh

from bfieldtools.mesh_conductor import MeshConductor
from bfieldtools.coil_optimize import optimize_streamfunctions
from bfieldtools.utils import combine_meshes, load_example_mesh


# Load simple plane mesh that is centered on the origin
planemesh = load_example_mesh("10x10_plane_hires")

# Specify coil plane geometry
center_offset = np.array([0, 0, 0])
standoff = np.array([0, 15, 0])

# Create coil plane pairs
coil_plus = trimesh.Trimesh(
    planemesh.vertices + center_offset + standoff, planemesh.faces, process=False
)

coil_minus = trimesh.Trimesh(
    planemesh.vertices + center_offset - standoff, planemesh.faces, process=False
)

joined_planes = combine_meshes((coil_plus, coil_minus))


# To spice things up, let's distort the planes a bit
joined_planes.vertices = (
    joined_planes.vertices
    - 0.5
    * np.linalg.norm(joined_planes.vertices, axis=1)[:, None]
    * np.sign(joined_planes.vertices[:, 1])[:, None]
    * joined_planes.vertex_normals
)

joined_planes.vertices = (
    joined_planes.vertices
    - 0.5
    * np.linalg.norm(joined_planes.vertices, axis=1)[:, None]
    * np.sign(joined_planes.vertices[:, 1])[:, None]
    * joined_planes.vertex_normals
)


# Create mesh class object
coil = MeshConductor(
    mesh_obj=joined_planes,
    fix_normals=True,
    basis_name="suh",
    N_suh=100,
    sph_radius=0.2,
    sph_normalization="energy",
)


#%%
# Set up target spherical harmonics components

target_alms = np.zeros((coil.opts["N_sph"] * (coil.opts["N_sph"] + 2),))
target_blms = np.zeros((coil.opts["N_sph"] * (coil.opts["N_sph"] + 2),))

target_blms[4] += 1


#%%
# Create bfield specifications used when optimizing the coil geometry


target_spec = {
    "coupling": coil.sph_couplings[1],
    "abs_error": 0.01,
    "target": target_blms,
}


#%%
# Run QP solver
import mosek

coil.s, prob = optimize_streamfunctions(
    coil,
    [target_spec],
    objective="minimum_ohmic_power",
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)

#%%
# Plot coil windings


f = coil.plot_mesh(opacity=0.2)

loops = coil.s.discretize(N_contours=6)

loops.plot_loops(figure=f)

#%%
# Now, let's change the spherical harmonics inner expansion radius (i.e. the target region radius)
# and optimize a new coil (with the same target sph component)

coil.set_sph_options(sph_radius=1.4)


target_spec = {
    "coupling": coil.sph_couplings[1],
    "abs_error": 0.01,
    "target": target_blms,
}


#%%
# Run QP solver
import mosek

coil.s2, prob = optimize_streamfunctions(
    coil,
    [target_spec],
    objective="minimum_ohmic_power",
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)

#%%
# Plot coil windings


f2 = coil.plot_mesh(opacity=0.2)

loops2 = coil.s2.discretize(N_contours=6)

loops2.plot_loops(figure=f2)



"""
Coil with interior holes
========================

Example showing a basic biplanar coil producing homogeneous field in a target
region between the two coil planes. The coil planes have holes in them,

"""

import numpy as np
import trimesh

from bfieldtools.mesh_conductor import MeshConductor
from bfieldtools.coil_optimize import optimize_streamfunctions
from bfieldtools.utils import combine_meshes, load_example_mesh


# Load simple plane mesh that is centered on the origin
planemesh = load_example_mesh("plane_w_holes")

angle = np.pi / 2
rotation_matrix = np.array(
    [
        [1, 0, 0, 0],
        [0, np.cos(angle), -np.sin(angle), 0],
        [0, np.sin(angle), np.cos(angle), 0],
        [0, 0, 0, 1],
    ]
)

planemesh.apply_transform(rotation_matrix)

# Specify coil plane geometry
center_offset = np.array([0, 0, 0])
standoff = np.array([0, 20, 0])

# Create coil plane pairs
coil_plus = trimesh.Trimesh(
    planemesh.vertices + center_offset + standoff, planemesh.faces, process=False
)

coil_minus = trimesh.Trimesh(
    planemesh.vertices + center_offset - standoff, planemesh.faces, process=False
)

joined_planes = combine_meshes((coil_plus, coil_minus))

# Create MeshConductor object, which finds the holes and sets the boundary condition
coil = MeshConductor(mesh_obj=joined_planes, fix_normals=True)

#%%
# Set up target and stray field points

# Here, the target points are on a volumetric grid within a sphere

center = np.array([0, 0, 0])

sidelength = 10
n = 8
xx = np.linspace(-sidelength / 2, sidelength / 2, n)
yy = np.linspace(-sidelength / 2, sidelength / 2, n)
zz = np.linspace(-sidelength / 2, sidelength / 2, n)
X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

x = X.ravel()
y = Y.ravel()
z = Z.ravel()

target_points = np.array([x, y, z]).T

# Turn cube into sphere by rejecting points "in the corners"
target_points = (
    target_points[np.linalg.norm(target_points, axis=1) < sidelength / 2] + center
)


#%%
# Create bfield specifications used when optimizing the coil geometry

# The absolute target field amplitude is not of importance,
# and it is scaled to match the C matrix in the optimization function

target_field = np.zeros(target_points.shape)
target_field[:, 0] = target_field[:, 0] + 1

target_abs_error = np.zeros_like(target_field)
target_abs_error[:, 0] += 0.001
target_abs_error[:, 1:3] += 0.005

target_spec = {
    "coupling": coil.B_coupling(target_points),
    "abs_error": target_abs_error,
    "target": target_field,
}

bfield_specification = [target_spec]

#%%
# Run QP solver
import mosek

coil.s, prob = optimize_streamfunctions(
    coil,
    bfield_specification,
    objective="minimum_inductive_energy",
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)

#%%
# Plot the computed streamfunction

coil.s.plot(ncolors=256)



"""
Gradient computation example
============================

Minimal example visualizing the gradient of a scalar function on a mesh

"""

import numpy as np
from mayavi import mlab
import trimesh

from bfieldtools.viz import plot_data_on_vertices
from bfieldtools.mesh_calculus import gradient
import pkg_resources

# Load simple plane mesh that is centered on the origin
file_obj = file_obj = pkg_resources.resource_filename(
    "bfieldtools", "example_meshes/10x10_plane.obj"
)
planemesh = trimesh.load(file_obj, process=False)

# Generate a simple scalar function
r = np.linalg.norm(planemesh.vertices, axis=1)
vals = np.exp(-0.5 * (r / r.max()))

# triangle centers for plotting
tri_centers = planemesh.vertices[planemesh.faces].mean(axis=1).T
tri_centers[1] += 0.1

#%%
#  Calculate the gradient (e.g., flow from potential)
g = gradient(vals, planemesh, rotated=False)

# Plot function and its gradient as arrows

scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))

plot_data_on_vertices(planemesh, vals, ncolors=15, figure=scene)
vecs = mlab.quiver3d(*tri_centers, *g, color=(1, 1, 1), mode="arrow", scale_factor=5)
vecs.glyph.glyph_source.glyph_position = "center"

#%%
# The same but rotated (e.g. current density from a stream function)
g = gradient(vals, planemesh, rotated=True)

# Plot function and its gradient as arrows
scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))

plot_data_on_vertices(planemesh, vals, ncolors=15, figure=scene)
vecs = mlab.quiver3d(*tri_centers, *g, color=(1, 1, 1), mode="arrow", scale_factor=5)
vecs.glyph.glyph_source.glyph_position = "center"



"""
Visualize SUH components on 3 surfaces
=====================================================
"""

import numpy as np
import matplotlib.pyplot as plt
from mayavi import mlab
import trimesh

from bfieldtools.suhtools import SuhBasis
from bfieldtools.utils import find_mesh_boundaries

from trimesh.creation import icosphere
from bfieldtools.utils import load_example_mesh

# Import meshes

# Sphere
sphere = icosphere(4, 0.1)

# Plane mesh centered on the origin
plane = load_example_mesh("10x10_plane_hires")
scaling_factor = 0.02
plane.apply_scale(scaling_factor)
# Rotate to x-plane
t = np.eye(4)
t[1:3, 1:3] = np.array([[0, 1], [-1, 0]])
plane.apply_transform(t)

# Bunny
bunny = load_example_mesh("bunny_repaired")
bunny.vertices -= bunny.vertices.mean(axis=0)

mlab.figure(bgcolor=(1, 1, 1))
for mesh in (sphere, plane, bunny):
    s = mlab.triangular_mesh(
        *mesh.vertices.T, mesh.faces, color=(0.5, 0.5, 0.5), opacity=0.2
    )
s.scene.z_plus_view()

Nc = 20
basis_sphere = SuhBasis(sphere, Nc)
basis_plane = SuhBasis(plane, Nc)
basis_bunny = SuhBasis(bunny, Nc)


A = [mesh.area for mesh in (bunny, sphere, plane)]

#%%

Nfuncs = [0, 1, 2, 3, 4, 5]
kwargs = {"colormap": "RdBu", "ncolors": 15}
plt.figure(figsize=(3.5, 2.5))
for i, b in enumerate((basis_bunny, basis_sphere, basis_plane)):
    fig = mlab.figure(bgcolor=(1, 1, 1), size=(550, 150))
    s = b.plot(Nfuncs, 0.1, Ncols=6, figure=fig, **kwargs)
    s[0].scene.parallel_projection = True
    s[0].scene.z_plus_view()
    if i == 0:
        s[0].scene.camera.parallel_scale = 0.1
    else:
        s[0].scene.camera.parallel_scale = 0.13
    plt.plot(np.sqrt(b.eigenvals * A[i]), ".-")

plt.legend(("bunny", "sphere", "square"), loc="lower right")
plt.xlabel("component index $n$")
plt.ylabel("$\sqrt{A}k_n$")
plt.tight_layout()



"""
Spherical harmonics B-field computation validation
==================================================



"""

import numpy as np
import trimesh
from mayavi import mlab

from bfieldtools.mesh_magnetics import magnetic_field_coupling
from bfieldtools.mesh_conductor import MeshConductor

from bfieldtools.sphtools import compute_sphcoeffs_mesh
from bfieldtools import sphtools


import pkg_resources

# Load simple plane mesh that is centered on the origin
file_obj = pkg_resources.resource_filename(
    "bfieldtools", "example_meshes/10x10_plane.obj"
)
coilmesh = trimesh.load(file_obj, process=False)
coil = MeshConductor(mesh_obj=coilmesh)

coil.mesh.vertices += np.array([0, -1, 0])
weights = np.zeros(coilmesh.vertices.shape[0])
weights[coil.inner_vertices] = 1

test_points = coilmesh.vertices.copy()
test_points[:, 1] = 0

lmax = 9


sph_C = compute_sphcoeffs_mesh(coil.mesh, lmax)

alms = sph_C[0] @ weights
blms = sph_C[1] @ weights

alms = np.zeros_like(alms)

B0 = (magnetic_field_coupling(coilmesh, test_points) @ weights).T
B1 = sphtools.field(test_points, alms, blms, lmax).T

mlab.figure(bgcolor=(1, 1, 1))

s = mlab.triangular_mesh(
    *coilmesh.vertices.T, coilmesh.faces, scalars=weights, colormap="viridis"
)
s.enable_contours = True
s.actor.property.render_lines_as_tubes = True
s.actor.property.line_width = 3.0

mlab.quiver3d(
    *test_points.T, *B0, color=(1, 0, 0), scale_factor=0.5e7, vmin=0, vmax=2e-7
)
mlab.quiver3d(
    *test_points.T, *B1, color=(0, 0, 1), scale_factor=0.5e7, vmin=0, vmax=2e-7
)
s.scene.isometric_view()


print(
    "Relative RMS error", np.sqrt(np.mean((B1 - B0) ** 2)) / np.sqrt(np.mean((B0) ** 2))
)



"""

Compare SUH and SPH basis functions for the magnetic scalar potential
======================================================================

"""


import numpy as np
import matplotlib.pyplot as plt
from mayavi import mlab
import trimesh
import pkg_resources
from pyface.api import GUI

_gui = GUI()


from bfieldtools.mesh_magnetics import magnetic_field_coupling
from bfieldtools.mesh_magnetics import magnetic_field_coupling_analytic
from bfieldtools.mesh_magnetics import scalar_potential_coupling
from bfieldtools.sphtools import compute_sphcoeffs_mesh, basis_fields
from bfieldtools.suhtools import SuhBasis
from bfieldtools.utils import load_example_mesh


mesh = load_example_mesh("bunny_repaired", process=True)

mesh.vertices -= mesh.vertices.mean(axis=0)

mesh_field = mesh.copy()
mesh_field.vertices += 0.001 * mesh_field.vertex_normals
# mesh_field = trimesh.smoothing.filter_laplacian(mesh_field, iterations=1)

# Ca, Cb = basis_fields(mesh_field.vertices, 4)

# bsuh = SuhBasis(mesh, 25)
# Csuh = magnetic_field_coupling_analytic(mesh, mesh_field.vertices) @ bsuh.basis

#%%
# def plot_basis_fields(C, comps):
#     d = 0.17
#     i = 0
#     j = 0
#     for n in comps:
#         p = 1.05 * mesh_field.vertices.copy()
#         p2 = mesh_field.vertices.copy()
#         #        p[:,1] -= i*d
#         #        p2[:,1] -= i*d
#         p[:, 0] += i * d
#         p2[:, 0] += i * d
#         m = np.max(np.linalg.norm(C[:, :, n], axis=0))
#         vectors = mlab.quiver3d(
#             *p.T, *C[:, :, n].T, mode="arrow", colormap="Greys", vmin=0, vmax=m
#         )
#         vectors.glyph.mask_input_points = True
#         vectors.glyph.mask_points.maximum_number_of_points = 1800
#         vectors.glyph.mask_points.random_mode_type = 1
#         vectors.glyph.glyph_source.glyph_position = "center"
#         vectors.glyph.glyph_source.glyph_source.shaft_radius = 0.02
#         vectors.glyph.glyph_source.glyph_source.tip_radius = 0.06
#         vectors.glyph.glyph.scale_factor = 0.03
#         #        m = np.max(abs((C[:,:,n].T*mesh_field.vertex_normals.T).sum(axis=0)))
#         #        s =mlab.triangular_mesh(*p.T, mesh_field.faces,
#         #                             scalars=(C[:,:,n].T*mesh_field.vertex_normals.T).sum(axis=0),
#         #                             colormap='seismic', vmin=-m, vmax=m, opacity=0.7)
#         #        s.actor.property.backface_culling = True
#         m = np.max(abs((C[:, :, n].T * mesh_field.vertex_normals.T).sum(axis=0)))
#         s = mlab.triangular_mesh(
#             *p2.T,
#             mesh.faces,
#             scalars=(C[:, :, n].T * mesh_field.vertex_normals.T).sum(axis=0),
#             colormap="bwr",
#             vmin=-m,
#             vmax=m
#         )
#         s.actor.mapper.interpolate_scalars_before_mapping = True
#         s.module_manager.scalar_lut_manager.number_of_colors = 15
#         i += 1


# comps = [0, 4, 10, 15]
# scene = mlab.figure(bgcolor=(1, 1, 1), size=(1200, 350))
# plot_basis_fields(Ca, comps)
# scene.scene.parallel_projection = True
# scene.scene.z_plus_view()
# scene.scene.camera.zoom(4)
# while scene.scene.light_manager is None:
#     _gui.process_events()
# scene.scene.light_manager.lights[2].intensity = 0.2


# scene = mlab.figure(bgcolor=(1, 1, 1), size=(1200, 350))
# plot_basis_fields(Csuh, comps)
# scene.scene.parallel_projection = True
# scene.scene.z_plus_view()
# scene.scene.camera.zoom(4)
# while scene.scene.light_manager is None:
#     _gui.process_events()
# scene.scene.light_manager.lights[2].intensity = 0.2


#%% test representation
from bfieldtools.mesh_magnetics import scalar_potential_coupling
from bfieldtools.sphtools import basis_potentials

scaling_factor = 0.02
# Load simple plane mesh that is centered on the origin
plane = load_example_mesh("10x10_plane_hires")
plane.apply_scale(scaling_factor)
# Rotate to x-plane
t = np.eye(4)
t[1:3, 1:3] = np.array([[0, 1], [-1, 0]])
plane.apply_transform(t)
# Subdivide face close to the mesh
from trimesh.proximity import signed_distance

dd = signed_distance(mesh, plane.triangles_center)
plane = plane.subdivide(np.flatnonzero(abs(dd) < 0.005))
dd = signed_distance(mesh, plane.triangles_center)
plane = plane.subdivide(np.flatnonzero(abs(dd) < 0.002))

UB = scalar_potential_coupling(mesh, plane.vertices, multiply_coeff=False)
mask = np.sum(UB, axis=1)
mask = mask > -0.5

Ua, Ub = basis_potentials(plane.vertices, 6)
bsuh = SuhBasis(mesh, 48)
# UB = scalar_potential_coupling(mesh, plane.vertices)
Usuh = UB @ bsuh.basis

sphere = trimesh.creation.icosphere(radius=0.02)
Ua_mesh, Ub_mesh = basis_potentials(sphere.vertices, 5)
UB_mesh = scalar_potential_coupling(mesh, mesh_field.vertices)
Usuh_mesh = UB_mesh @ bsuh.basis


#%%
comps = [0, 5, 13, 17]
d = 0.22
from bfieldtools.viz import plot_data_on_vertices
from bfieldtools.viz import plot_mesh

# Plot suh
i = 0
fig = mlab.figure(bgcolor=(1, 1, 1), size=(600, 190))
for n in comps:
    p = plane.copy()
    p.vertices[:, 0] += i * d
    p2 = mesh_field.copy()
    p2.vertices[:, 0] += i * d
    scalars = Usuh[:, n].copy()
    scalars[~mask] = 0
    scalars2 = bsuh.basis[:, n]
    m = np.max(abs(scalars))
    m2 = np.max(abs(scalars2))
    plot_data_on_vertices(p, scalars, figure=fig, ncolors=15, vmax=m, colormap="bwr")
    plot_data_on_vertices(
        p2, scalars2, figure=fig, ncolors=15, vmax=m2, colormap="BrBG"
    )
    i += 1
fig.scene.parallel_projection = True
fig.scene.z_plus_view()
fig.scene.camera.parallel_scale = 0.11

#%% Plot SPH
i = 0
fig = mlab.figure(bgcolor=(1, 1, 1), size=(600, 190))
for n in comps:
    p = plane.copy()
    p.vertices[:, 0] += i * d
    p2 = sphere.copy()
    p2.vertices[:, 0] += i * d
    scalars = Ua[:, n].copy()
    scalars[~mask] = 0
    scalars2 = Ua_mesh[:, n]
    m = np.max(abs(scalars))
    m2 = np.max(abs(scalars2))
    plot_data_on_vertices(p, scalars, figure=fig, ncolors=15, vmax=m, colormap="bwr")
    plot_data_on_vertices(
        p2, scalars2, figure=fig, ncolors=15, vmax=m2, colormap="BrBG"
    )
    p3 = mesh.copy()
    p3.vertices[:, 0] += i * d
    plot_mesh(p3, opacity=0.3, figure=fig)
    i += 1
fig.scene.parallel_projection = True
fig.scene.z_plus_view()
fig.scene.camera.parallel_scale = 0.11



"""
Uniform triangle
================

Test and validation of potential of uniformaly distributed charge density

For the math see:
        A. S. Ferguson, Xu Zhang and G. Stroink,
        "A complete linear discretization for calculating the magnetic field
        using the boundary element method,"
        in IEEE Transactions on Biomedical Engineering,
        vol. 41, no. 5, pp. 455-460, May 1994.
        doi: 10.1109/10.293220

"""

import numpy as np
import matplotlib.pyplot as plt
import sys
from mayavi import mlab

path = "/m/home/home8/80/makinea1/unix/pythonstuff/bfieldtools"
if path not in sys.path:
    sys.path.insert(0, path)

from bfieldtools.integrals import triangle_potential_uniform
from bfieldtools.utils import tri_normals_and_areas

#%%
#%% Test potential shape slightly above the surface
points = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]])

tris = np.array([[0, 1, 2]])
p_tris = points[tris]

# Evaluation points
Nx = 100
xx = np.linspace(-2, 2, Nx)
X, Y = np.meshgrid(xx, xx, indexing="ij")
Z = np.zeros_like(X) + 0.01
p_eval = np.array([X, Y, Z]).reshape(3, -1).T

# Difference vectors
RR = p_eval[:, None, None, :] - p_tris[None, :, :, :]
tn, ta = tri_normals_and_areas(points, tris)

pot = triangle_potential_uniform(RR, tn, False)

# Plot shape
plt.figure()
plt.imshow(pot[:, 0].reshape(Nx, Nx), extent=(xx.min(), xx.max(), xx.max(), xx.min()))
plt.ylabel("x")
plt.xlabel("y")

#%%
#%% Test asymptotic behavour by comparing potential of  charge at the
#    center of mass of the triangle having the same first moment
def charge_potential(Reval, Rcharge, moment):
    R = Reval - Rcharge
    r = np.linalg.norm(R, axis=1)
    return moment / r


# Center of mass
Rcharge = points.mean(axis=0)
# Moment
m = ta
# Eval points
Neval = 100

cases = np.arange(3)
f, ax = plt.subplots(1, 3)
mlab.figure(bgcolor=(1, 1, 1))
mlab.triangular_mesh(*points.T, tris, color=(0.5, 0.5, 0.5))
for c in cases:
    p_eval2 = np.zeros((Neval, 3))

    if c == 0:
        z = np.linspace(0.01, 20, Neval)
        p_eval2[:, 2] = z
        p_eval2 += Rcharge
        mlab.points3d(*p_eval2.T, color=(1, 0, 0), scale_factor=0.1)
        lab = "z"
    elif c == 1:
        x = np.linspace(0.01, 20, Neval)
        p_eval2[:, 0] = x
        mlab.points3d(*p_eval2.T, color=(0, 1, 0), scale_factor=0.1)
        lab = "x"
    elif c == 2:
        y = np.linspace(0.01, 20, Neval)
        p_eval2[:, 1] = y
        mlab.points3d(*p_eval2.T, color=(0, 0, 1), scale_factor=0.1)
        lab = "y"

    plt.sca(ax[c])
    # Plot dipole field approximating uniform dipolar density
    plt.semilogy(z, charge_potential(p_eval2, Rcharge, m))
    # Plot sum of the linear dipoles
    RR = p_eval2[:, None, None, :] - p_tris[None, :, :, :]
    pot = triangle_potential_uniform(RR, tn, False)
    plt.semilogy(z, pot)
    plt.xlabel(lab)
    if c == 0:
        plt.ylabel("potential")
    if c == 2:
        plt.legend(("Approx.", "True"))



"""
Field of a Helmholtz coils pair 
==========================================================

Example on how to compute the magnetic field from current line segments forming a Helmholtz coil pair.

Visualization of the 3D field using Mayavi
"""

from mayavi import mlab
import matplotlib.pyplot as plt
import numpy as np

from bfieldtools.utils import cylinder_points
from bfieldtools.line_magnetics import magnetic_field

# Create helmholtz coil with radius R
R = 5

c_points = cylinder_points(
    radius=R, length=0, nlength=1, nalpha=100, orientation=np.array([0, 1, 0])
)
c_points[:, 1] = 0
c_points = np.vstack((c_points, c_points[0, :]))


c1_points = c_points - np.array([0, R / 2, 0])
c2_points = c_points + np.array([0, R / 2, 0])

mlab.figure(bgcolor=(1, 1, 1))
mlab.plot3d(*c1_points.T)
mlab.plot3d(*c2_points.T)

box = 3 * R
n = 50

xx = np.linspace(-box, box, n)
yy = np.linspace(-box, box, n)
zz = np.linspace(-box, box, n)
X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

x = X.ravel()
y = Y.ravel()
z = Z.ravel()

b_points = np.array([x, y, z]).T

B = np.zeros(b_points.shape)

B += magnetic_field(c1_points, b_points)
B += magnetic_field(c2_points, b_points)


B_matrix = B.reshape((n, n, n, 3))

B_matrix_norm = np.linalg.norm(B_matrix, axis=-1)

#%% Visualization using mayavi

mlab.figure(bgcolor=(1, 1, 1))

field = mlab.pipeline.vector_field(
    X,
    Y,
    Z,
    B_matrix[:, :, :, 0],
    B_matrix[:, :, :, 1],
    B_matrix[:, :, :, 2],
    scalars=B_matrix_norm,
    name="B-field",
)

vectors = mlab.pipeline.vectors(field, scale_factor=(X[1, 0, 0] - X[0, 0, 0]),)


vectors.glyph.mask_input_points = True
vectors.glyph.mask_points.on_ratio = 2

vcp = mlab.pipeline.vector_cut_plane(field)
vcp.glyph.glyph.scale_factor = 10 * (X[1, 0, 0] - X[0, 0, 0])
# For prettier picture:
vcp.implicit_plane.widget.enabled = True

iso = mlab.pipeline.iso_surface(field, contours=10, opacity=0.2, colormap="viridis")


# A trick to make transparency look better: cull the front face
iso.actor.property.frontface_culling = True

# Settings
iso.contour.maximum_contour = 1e-07
vcp.implicit_plane.widget.normal_to_y_axis = True


#%% Plot field magnitude on x and y axes using matplotlib

plt.figure()

z1 = np.linspace(0, 30, 31)

x1 = y1 = np.zeros_like(z1)

line1_points = np.vstack((x1, y1, z1)).T


Bh_line1 = magnetic_field(c1_points, line1_points) + magnetic_field(
    c2_points, line1_points
)

plt.semilogy(
    z1,
    np.linalg.norm(Bh_line1, axis=1) / np.linalg.norm(Bh_line1, axis=1)[0],
    label="Helmholtz, z-axis",
)


y2 = np.linspace(0, 30, 31)

z2 = x2 = np.zeros_like(y2)

line2_points = np.vstack((x2, y2, z2)).T

Bh_line2 = magnetic_field(c1_points, line2_points) + magnetic_field(
    c2_points, line2_points
)


plt.semilogy(
    y2,
    np.linalg.norm(Bh_line2, axis=1) / np.linalg.norm(Bh_line2, axis=1)[0],
    label="Helmholtz, y-axis",
)
plt.ylabel("Field amplitude (target field units)")
plt.xlabel("Distance from origin")
plt.grid(True, which="minor", axis="y")
plt.grid(True, which="major", axis="y", color="k")
plt.grid(True, which="major", axis="x")

plt.legend()



"""
Head gradient coil
==================

Example showing a gradient coil designed on the surface of a MEG system helmet
"""


import numpy as np
from mayavi import mlab

from bfieldtools.mesh_conductor import MeshConductor
from bfieldtools.coil_optimize import optimize_streamfunctions
from bfieldtools.utils import load_example_mesh
from bfieldtools import sphtools


# Load simple plane mesh that is centered on the origin
helmetmesh = load_example_mesh("meg_helmet")
coil = MeshConductor(mesh_obj=helmetmesh, fix_normals=True)

#%%
# Set up target and stray field points.
# Here, the target points are on a volumetric grid within a sphere

offset = np.array([0, 0, 0.04])
center = offset

sidelength = 0.05
n = 12
xx = np.linspace(-sidelength / 2, sidelength / 2, n)
yy = np.linspace(-sidelength / 2, sidelength / 2, n)
zz = np.linspace(-sidelength / 2, sidelength / 2, n)
X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

x = X.ravel()
y = Y.ravel()
z = Z.ravel()

target_points = np.array([x, y, z]).T

# Turn cube into sphere by rejecting points "in the corners"
# and inner points
target_points = (
    target_points[
        (np.linalg.norm(target_points, axis=1) < sidelength / 2)
        * (np.linalg.norm(target_points, axis=1) > sidelength / 2 * 0.8)
    ]
    + center
)


#%%
# Specify target field and run solver.
# Here, we specify the target field through the use of spherical harmonics.
# We want to produce the field corresponding to a specific beta_l,m-component.

lmax = 3
alm = np.zeros((lmax * (lmax + 2),))
blm = np.zeros((lmax * (lmax + 2),))

# Set one specific component to one
blm[3] += 1

sphfield = sphtools.field(target_points, alm, blm, lmax)

target_field = sphfield / np.max(sphfield[:, 0])

target_field[:, 2] = 0

coil.plot_mesh(opacity=0.5)
mlab.quiver3d(*target_points.T, *sphfield.T)
mlab.gcf().scene.isometric_view()

abs_error = np.zeros_like(target_field)
abs_error[:, 0] += 0.05
abs_error[:, 1:3] += 0.1


target_spec = {
    "coupling": coil.B_coupling(target_points),
    "abs_error": abs_error,
    "target": target_field,
}

#%%
import mosek

coil.s, prob = optimize_streamfunctions(
    coil,
    [target_spec],
    objective="minimum_inductive_energy",
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)

#%%
# Plot coil windings


loops = coil.s.discretize(N_contours=10)
loops.plot_loops()



import time
import numpy as np

from ..mesh_calculus import gradient_matrix
from .integrals import omega, gamma0


def magnetic_field_coupling_analytic(mesh, r, Nchunks=None):
    """
    **DEPRECATED**
    Given a mesh, computes the "C matrix" which gives the magnetic field at
    some target points due to currents (stream function) on a surface mesh.

    Parameters
    ----------

    mesh: Trimesh mesh object
            mesh describing the geometry of the field source
    r: ndarray (Np, 3)
        evaluation points

    Returns
    -------
    C: ndarray (Np, 3, Nvertices)
        Coupling matrix corresponding to a mapping from a stream function
        on the mesh to B-field at the evaluation points

    **DEPRECATED**
    """

    coef = 1e-7

    print(
        "Computing magnetic field coupling matrix analytically, %d vertices by %d target points... "
        % (len(mesh.vertices), len(r)),
        end="",
    )
    start = time.time()

    if Nchunks is None:
        if r.shape[0] > 1000:
            Nchunks = r.shape[0] // 100
        else:
            Nchunks = 1

    #    ta = mesh.area_faces
    tn = mesh.face_normals

    # Nfaces, 3, 3
    rfaces = mesh.vertices[mesh.faces]
    # Neval, Nfaces, xyz
    coeffs = np.zeros((r.shape[0:1] + mesh.faces.shape[:-1] + (3,)))
    # Nfaces, Nedges, 3
    edges = np.roll(rfaces, 1, -2) - np.roll(rfaces, 2, -2)
    #    grad = np.cross(tn[:, None, :], edges, axis=-1)/(2*ta[:, None, None])
    Gx, Gy, Gz = gradient_matrix(mesh, rotated=False)
    Rx, Ry, Rz = gradient_matrix(mesh, rotated=True)
    solid_angle = np.zeros((r.shape[0], rfaces.shape[0]))
    # Calculate potentials and related coefficients
    for n in range(Nchunks):
        RRchunk = r[n::Nchunks, None, None, :] - rfaces[None, :, :, :]
        # Neval, Nfaces, Nedges
        #        result = -np.sum(np.sum(gamma0(RRchunk)[..., None]*edges,
        #                               axis=-2)[...,None, :]*edges, axis=-1)
        #        result *= 1/(2*ta[..., :, None])
        solid_angle[n::Nchunks] = omega(RRchunk)
        coeffs[n::Nchunks] = -np.einsum("...i,...ik->...k", gamma0(RRchunk), edges)

    #    # Accumulate the elements
    C = np.zeros((3, r.shape[0], mesh.vertices.shape[0]))
    #    for ind_f, f in enumerate(mesh.faces):
    #        C[:, f, :] += coeffs[:, ind_f, :, None]*tn[ind_f]
    #        C[:, f, :] += -solid_angle[:, ind_f:ind_f+1, None]*grad[ind_f]

    G = (Gx, Gy, Gz)
    for i in range(3):
        cc = coeffs * tn[:, i, None]
        C[i] += cc[:, :, 0] @ Rx + cc[:, :, 1] @ Ry + cc[:, :, 2] @ Rz
        C[i] += -solid_angle @ G[i]

    duration = time.time() - start
    print("took %.2f seconds." % duration)

    C *= coef
    #    return np.moveaxis(C, 0, 1)
    return np.swapaxes(C, 0, 1)
